from typing import assert_never, cast

from rslearn.config import dataset
from rslearn.dataset.window import get_window_layer_dir
from upath import UPath

from olmoearth_run.runner.models import rslearn_config
from olmoearth_run.runner.models.operational_context import ModelOperationalContext
from olmoearth_run.runner.tools.olmoearth_config.constants import OLMOEARTH_SENTINEL2_L2A_BANDS, LABEL_LAYER_NAME, SENTINEL2_L2A, DATA_SPLIT_KEY, TARGET, TARGET_SELECTOR_PREFIX, OUTPUT_LAYER_NAME
from olmoearth_run.shared.models.api.step_type import StepType
from olmoearth_run.shared.models.data_split_type import DataSplitType
from olmoearth_shared.models.olmoearth_config.data.data import Data, Output
from olmoearth_shared.models.olmoearth_config.data.output import RegressionField, SegmentationField, RasterOutput, VectorOutput, RasterField
from olmoearth_shared.models.olmoearth_config.data.temporality import Temporality, SampledTemporality, RepeatingIntervalTemporality
from olmoearth_shared.models.olmoearth_config.data_type import DataType
from olmoearth_shared.models.olmoearth_config.input_preprocessing.input_preprocessing import DefaultInputPreprocessing, TrainingSplitInputPreprocessing, PredictInputPreprocessing
from olmoearth_shared.models.olmoearth_config.input_preprocessing.transforms import Transform, OlmoEarthNormalize, RandomFlip
from olmoearth_shared.models.olmoearth_config.labeled_data_prep.labeled_data_prep import LabeledDataPrep
from olmoearth_shared.models.olmoearth_config.labeled_data_prep.window_preparers import PointToRasterWindowPreparer, PolygonToRasterWindowPreparer
from olmoearth_shared.models.olmoearth_config.model.encoders import Encoder, EscapeHatchEncoder, OlmoEarthEncoder
from olmoearth_shared.models.olmoearth_config.model.model import Model
from olmoearth_shared.models.olmoearth_config.input_preprocessing.transforms import TransformName
from olmoearth_shared.models.olmoearth_config.model.tasks import SegmentationTask
from olmoearth_shared.models.olmoearth_config.olmoearth_config import OlmoEarthConfig


class DataTranspiler:
    @staticmethod
    def generate_data_module_config(olmoearth_config: OlmoEarthConfig, ops_context: ModelOperationalContext) -> rslearn_config.DataModule:
        model = olmoearth_config.model
        preprocessing = olmoearth_config.input_preprocessing
        data = olmoearth_config.data

        if not model:
            raise ValueError("No model configuration provided, cannot make a data module config")

        if not preprocessing:
            raise ValueError("No input preprocessing configuration provided, cannot make a data module config")

        if not data.output:
            raise ValueError("No output schema specified, cannot make a data module config")

        return rslearn_config.RslearnDataModule(
            class_path=rslearn_config.ClassPath.RSLEARN_DATA_MODULE,
            init_args=rslearn_config.RslearnDataModuleInitArgs(
                path=ops_context.dataset_path,
                batch_size=OpsUtils.get_batch_size(olmoearth_config, ops_context),
                num_workers=ops_context.num_data_worker_processes,
                inputs=InputsTranspilers.generate_inputs(olmoearth_config.data, model.encoder, ops_context),
                task=TaskTranspilers.generate_task(model, data.output),
                default_config=PreprocessingTranspilers.generate_default_config(preprocessing.default, model, olmoearth_config.data),
                train_config=PreprocessingTranspilers.generate_training_split_config(DataSplitType.TRAIN, preprocessing.default, preprocessing.train, model, olmoearth_config.data),
                val_config=PreprocessingTranspilers.generate_training_split_config(DataSplitType.VAL, preprocessing.default, preprocessing.val, model, olmoearth_config.data),
                test_config=PreprocessingTranspilers.generate_training_split_config(DataSplitType.TEST, preprocessing.default, preprocessing.test, model, olmoearth_config.data),
                predict_config=PreprocessingTranspilers.generate_predict_config(
                    preprocessing.default,
                    preprocessing.predict,
                    model,
                    olmoearth_config.data,
                    predict_groups=ops_context.predict_groups,
                    step_type=ops_context.step_type,
                    labeled_data_prep=olmoearth_config.labeled_data_prep,
                ),
            ),
        )


class InputsTranspilers:
    @staticmethod
    def generate_inputs(data: Data, encoder: Encoder, ops_context: ModelOperationalContext) -> dict[str, rslearn_config.DataInputArgs]:
        inputs: dict[str, rslearn_config.DataInputArgs] = {}

        modalities = data.modalities
        if modalities.sentinel2_l2a:
            inputs[SENTINEL2_L2A] = InputsTranspilers.generate_sentinel2_l2a_input_args(data.temporality, encoder)

        if ops_context.needs_targets:
            if data.output is None:
                raise ValueError("No output schema specified, cannot generate targets")
            inputs[TARGET] = InputsTranspilers.generate_training_target_args(data.output)

        return inputs

    @staticmethod
    def generate_sentinel2_l2a_input_args(temporality: Temporality, encoder: Encoder) -> rslearn_config.DataInputArgs:
        match encoder:
            case OlmoEarthEncoder():
                return rslearn_config.DataInputArgs(
                    data_type=DataType.RASTER,
                    layers=InputsTranspilers.generate_layer_names(temporality, SENTINEL2_L2A),
                    bands=OLMOEARTH_SENTINEL2_L2A_BANDS,
                    passthrough=True,
                    dtype=dataset.DType.FLOAT32.name,
                    load_all_layers=True,
                )
            case EscapeHatchEncoder():
                raise NotImplementedError("Escape hatch encoders are not supported at this time")
            case _:
                assert_never(encoder)

    @staticmethod
    def generate_training_target_args(output: Output) -> rslearn_config.DataInputArgs:
        if len(output.fields) != 1:
            raise NotImplementedError("Only one output field per model is supported at this time")

        field_name = list(output.fields.keys())[0]
        field = list(output.fields.values())[0]

        match output:
            case RasterOutput():
                return InputsTranspilers.generate_raster_training_target_args(field_name, cast(RasterField, field))
            case VectorOutput():
                raise NotImplementedError("Vector outputs are not supported at this time")
            case _ as unreachable:
                assert_never(unreachable)

    @staticmethod
    def generate_raster_training_target_args(field_name: str, field: RasterField) -> rslearn_config.DataInputArgs:
        match field:
            case SegmentationField():
                dtype = dataset.DType.INT32.name
            case RegressionField():
                dtype = dataset.DType.FLOAT32.name
            case _ as unreachable:
                assert_never(unreachable)

        return rslearn_config.DataInputArgs(
            data_type=DataType.RASTER,
            layers=[LABEL_LAYER_NAME],
            bands=[field_name],
            dtype=dtype,
            is_target=True,
        )

    @staticmethod
    def generate_layer_names(temporality: Temporality, modality_name: str) -> list[str]:
        # TODO: DRY up this temporality-> num images logic. used in dataset generation too.
        match temporality:
            case SampledTemporality():
                num_layers = temporality.num_samples
            case RepeatingIntervalTemporality():
                num_layers = temporality.num_periods
            case _:
                assert_never(temporality)

        # Use a dummy window path - we only care about the folder name at the end
        dummy_window_path = UPath("/dummy")

        layer_names = []
        for group_idx in range(num_layers):
            layer_dir = get_window_layer_dir(dummy_window_path, modality_name, group_idx)
            layer_name = layer_dir.name
            layer_names.append(layer_name)

        return layer_names


class TaskTranspilers:
    @staticmethod
    def generate_task(model: Model, output: Output) -> rslearn_config.Task:
        if len(model.tasks) != 1:
            raise NotImplementedError("Only one task is supported for now")

        task = list(model.tasks.values())[0]
        task_name = list(model.tasks.keys())[0]

        match task:
            case SegmentationTask():
                return TaskTranspilers.generate_segmentation_task(task_name, task, output)
            case _ as unreachable:
                assert_never(unreachable)

    @staticmethod
    def generate_segmentation_task(task_name: str, task: SegmentationTask, output: Output) -> rslearn_config.Task:
        field = output.fields[task_name]
        if not isinstance(field, SegmentationField):
            raise ValueError(f"Task {task_name} is not a segmentation task")

        num_classes = len(field.allowed_values)
        nodata_value = field.nodata_value

        return rslearn_config.SegmentationTask(
            class_path=rslearn_config.ClassPath.SEGMENTATION_TASK,
            init_args=rslearn_config.SegmentationTaskInitArgs(
                num_classes=num_classes,
                nodata_value=nodata_value,
                metric_kwargs={"average": "micro"},
            )
        )


class PreprocessingTranspilers:
    @staticmethod
    def generate_default_config(default: DefaultInputPreprocessing, model: Model, data: Data) -> rslearn_config.PreprocessingArgs:
        return rslearn_config.PreprocessingArgs(
            patch_size=default.input_size,
            transforms=PreprocessingTranspilers.generate_transforms_list(default.transforms, None, None, model.encoder, data),
        )

    @staticmethod
    def generate_training_split_config(split_type: DataSplitType, default: DefaultInputPreprocessing, training_split: TrainingSplitInputPreprocessing | None, model: Model, data: Data) -> rslearn_config.PreprocessingArgs:
        tags = {DATA_SPLIT_KEY: split_type.value}
        transforms_list = None
        patch_size = None

        if training_split is not None:
            transforms_list = PreprocessingTranspilers.generate_transforms_list(default.transforms, training_split.transforms, training_split.additional_transforms, model.encoder, data)
            patch_size = training_split.input_size if training_split.input_size is not None else None

            if training_split.sampler is not None:
                raise NotImplementedError("Samplers are not yet supported at this time")

        return rslearn_config.PreprocessingArgs(
            tags=tags,
            patch_size=patch_size,
            transforms=transforms_list,
        )

    @staticmethod
    def generate_predict_config(
        default: DefaultInputPreprocessing,
        predict_split: PredictInputPreprocessing | None,
        model: Model,
        data: Data,
        predict_groups: list[str] | None = None,
        step_type: StepType | None = None,
        labeled_data_prep: LabeledDataPrep | None = None,
    ) -> rslearn_config.PreprocessingArgs:
        input_size = default.input_size
        overlap_pixels = 0
        transforms_list = None
        patch_size = None
        tags = None

        if step_type == StepType.MODEL_EVALUATION:
            # Match training validation behavior for evaluation
            tags = {DATA_SPLIT_KEY: DataSplitType.VAL.value}
            skip_targets = False

            if not labeled_data_prep:
                raise ValueError(f"Cannot perform step {StepType.MODEL_EVALUATION} without labeled_data_prep config")

            # For point annotations, use single crop per window to match validation distribution
            # For polygon annotations, use full coverage since the whole window contains labels
            match labeled_data_prep.window_preparer:
                case PointToRasterWindowPreparer():
                    load_all_patches = False
                case PolygonToRasterWindowPreparer():
                    load_all_patches = True
                case _ as unreachable:
                    assert_never(unreachable)
        else:
            # For user predictions: full coverage, skip targets
            load_all_patches = True
            skip_targets = True

        if predict_split is not None:
            if predict_split.input_size is not None:
                input_size = predict_split.input_size
                patch_size = predict_split.input_size
            overlap_pixels = predict_split.overlap_pixels

            if predict_split.transforms is not None or predict_split.additional_transforms is not None:
                transforms_list = PreprocessingTranspilers.generate_transforms_list(
                    default.transforms,
                    predict_split.transforms,
                    predict_split.additional_transforms,
                    model.encoder,
                    data
                )

        overlap_ratio = None if overlap_pixels == 0 else overlap_pixels / input_size

        return rslearn_config.PreprocessingArgs(
            tags=tags,
            patch_size=patch_size,
            transforms=transforms_list,
            load_all_patches=load_all_patches,
            skip_targets=skip_targets,
            overlap_ratio=overlap_ratio,
            groups=predict_groups,
            output_layer_name_skip_inference_if_exists=OUTPUT_LAYER_NAME,
        )

    @staticmethod
    def generate_transforms_list(default_transforms: list[Transform], override_transforms: list[Transform] | None, additional_transforms: list[Transform] | None, encoder: Encoder, data: Data) -> list[rslearn_config.Transform]:
        transforms: list[Transform] = []

        match encoder:
            case OlmoEarthEncoder():
                transforms.append(OlmoEarthNormalize(name=TransformName.OLMOEARTH_NORMALIZE))
            case EscapeHatchEncoder():
                raise NotImplementedError("Escape hatch encoders are not supported at this time")
            case _:
                assert_never(encoder)

        if override_transforms is not None:
            transforms.extend(override_transforms)
        elif additional_transforms is not None:
            transforms.extend(default_transforms)
            transforms.extend(additional_transforms)
        else:
            transforms.extend(default_transforms)

        return [TransformTranspilers.generate_transform(transform, data) for transform in transforms]


class TransformTranspilers:
    @staticmethod
    def generate_transform(transform: Transform, data: Data) -> rslearn_config.Transform:
        match transform:
            case OlmoEarthNormalize():
                return TransformTranspilers.generate_olmoearth_normalize_transform(data)
            case RandomFlip():
                return TransformTranspilers.generate_random_flip_transform(transform, data)
            case _ as unreachable:
                assert_never(unreachable)

    @staticmethod
    def generate_olmoearth_normalize_transform(data: Data) -> rslearn_config.Transform:
        band_names = {}
        if data.modalities.sentinel2_l2a:
            band_names[SENTINEL2_L2A] = OLMOEARTH_SENTINEL2_L2A_BANDS

        return rslearn_config.OlmoEarthNormalize(
            class_path=rslearn_config.ClassPath.OLMOEARTH_NORMALIZE,
            init_args=rslearn_config.OlmoEarthNormalizeInitArgs(band_names=band_names),
        )

    @staticmethod
    def generate_random_flip_transform(transform: RandomFlip, data: Data) -> rslearn_config.Transform:
        modalities = data.modalities

        image_selectors: list[str] = []
        if modalities.sentinel2_l2a:
            image_selectors.append(SENTINEL2_L2A)
        if data.output is not None:
            match data.output:
                case RasterOutput():
                    if len(data.output.fields) != 1:
                        raise NotImplementedError("Only one output field per model is supported at this time")
                    field = list(data.output.fields.values())[0]
                    match field:
                        case SegmentationField():
                            image_selectors.append(f"{TARGET_SELECTOR_PREFIX}/classes")
                        case RegressionField():
                            image_selectors.append(f"{TARGET_SELECTOR_PREFIX}/values")
                        case _ as unreachable:
                            assert_never(unreachable)
                    image_selectors.append(f"{TARGET_SELECTOR_PREFIX}/valid")

                case VectorOutput():
                    raise NotImplementedError("Vector outputs are not supported at this time")

                case _:
                    assert_never(data.output)

        return rslearn_config.Flip(
            class_path=rslearn_config.ClassPath.RANDOM_FLIP,
            init_args=rslearn_config.FlipInitArgs(
                image_selectors=image_selectors,
                horizontal=transform.x,
                vertical=transform.y,
            ),
        )


class OpsUtils:
    @staticmethod
    def get_batch_size(olmoearth_config: OlmoEarthConfig, ops_context: ModelOperationalContext) -> int:
        if not ops_context.is_training:
            return ops_context.batch_size

        training = olmoearth_config.training
        if not training:
            raise ValueError("No training configuration provided, cannot get batch size")

        return training.batch_size



from enum import StrEnum


class DataSplitType(StrEnum):
    """
    An enumerated type for data splits used in machine learning training.
    """
    TRAIN = "train"
    VAL = "val"
    TEST = "test"



from upath import UPath

from olmoearth_shared.api.run.prediction_geometry import PredictionRequestFeature, PredictionResultCollection


class PostprocessInterfaceVector:
    """
    Interface describing how we post-process the vector results of either a Task or an entire Workflow.
    """

    def process_window(self, window_request: PredictionRequestFeature, window_result: PredictionResultCollection) -> PredictionResultCollection:
        """
        In the context of a single Window, combine the request geojson used to initialize the Window with the FeatureCollection generated by inference
        """
        raise NotImplementedError

    def process_partition(self, partition_window_results: list[PredictionResultCollection]) -> PredictionResultCollection:
        """
        Combine the results of process_window() for all windows in the partition, resulting in a single FeatureCollection for the partition
        """
        raise NotImplementedError

    def process_dataset(self, all_partitions_results: list[PredictionResultCollection]) -> PredictionResultCollection:
        """
        Combine the results of all partitions into a single FeatureCollection for the entire prediction request
        """
        raise NotImplementedError


class PostprocessInterfaceRaster:
    """
    Interface describing how we post-process the raster results of either a Task or an entire Workflow.
    """

    def process_window(self, window_request: PredictionRequestFeature, window_result_path: UPath) -> None:
        """
        In the context of a single Window, combine the request geojson used to initialize the Window with the FeatureCollection generated by inference.
        This might involve setting values in the GeoTIFF properties, applying a vector mask to the raster, and so on.
        """
        raise NotImplementedError

    def process_partition(self, partition_window_results: list[UPath], output_dir: UPath) -> list[UPath]:
        """
        Combine the GeoTIFFs from each window into a single geotiff for the whole partition.
        """
        raise NotImplementedError

    def process_dataset(self, all_partitions_result_paths: list[UPath], output_dir: UPath) -> list[UPath]:
        """
        Combine the GeoTIFFs from each window into a single geotiff for the whole partition.
        """
        raise NotImplementedError



from enum import StrEnum
from typing import Annotated, Literal

from pydantic import BaseModel, Field
from typing_extensions import TypeAliasType


class SegmentationDecoderName(StrEnum):
    UPSAMPLE_CONV = "upsample_conv"
    UNET = "unet"


class UpsampleConvSegmentationDecoder(BaseModel):
    """
    A decoder that upsamples the embeddings using a convolutional layer before passing to a segmentation decoder head.
    """
    name: Literal[SegmentationDecoderName.UPSAMPLE_CONV]


class UnetSegmentationDecoder(BaseModel):
    """
    A decoder that uses a U-Net architecture to perform segmentation.
    """
    name: Literal[SegmentationDecoderName.UNET]


SegmentationDecoder = TypeAliasType(
    "SegmentationDecoder",
    Annotated[
        UpsampleConvSegmentationDecoder | UnetSegmentationDecoder,
        Field(discriminator="name")
    ]
)


# TODO: classification, per-pixel regression, regression, and detection decoders



import logging
import shutil
import uuid
from pathlib import Path
from os import PathLike

from olmoearth_run.runner.steps.prepare_labeled_windows_step_definition import (
    PrepareLabeledWindowsStepDefinition,
)
from olmoearth_run.shared.models.model_stage_paths import (
    ANNOTATION_FEATURES_FILE_NAME,
    ANNOTATION_TASK_FEATURES_FILE_NAME,
)
from olmoearth_run.shared.models.fine_tuning_scratch_space import (
    FineTuningScratchSpace,
    OLMOEARTH_CONFIG_FILE_NAME,
)
from olmoearth_run.shared.models.api.task_args import (
    PrepareLabeledWindowsTaskArgs,
    DatasetBuildFromWindowsTaskArgs,
    FineTuneTaskArgs,
)
from olmoearth_run.shared.models.api.task_results import (
    PrepareLabeledWindowsTaskResults,
    DatasetBuildFromWindowsTaskResults,
    FineTuneTaskResults,
)
from olmoearth_run.runner.steps.dataset_build_from_windows_step_definition import (
    DatasetBuildFromWindowsStepDefinition,
)
from olmoearth_run.runner.steps.fine_tuning_step_definition import (
    FineTuningStepDefinition,
)


logger = logging.getLogger(__name__)


class UnifiedConfigFineTuneRunner:
    """
    Local runner for the OlmoEarth fine-tuning pipeline.

    Uses the unified OlmoEarthConfig format (olmoearth_config.yaml) for configuration.
    """

    def __init__(
        self,
        project_path: PathLike,
        scratch_path: PathLike,
    ):
        self.project_path = Path(project_path)
        self.scratch_path = Path(scratch_path)
        self.scratch_space = FineTuningScratchSpace(root_path=str(self.scratch_path))
        self.model_stage_paths = self.scratch_space.model_stage_paths
        self.dataset_path = Path(self.model_stage_paths.default_dataset_path)
        self.step_id = uuid.uuid4()

        self._setup_project_env()

    def _setup_project_env(self) -> None:
        """Set up the project environment in the scratch space by copying config and creating symlinks."""

        self.dataset_path.mkdir(parents=True, exist_ok=True)

        # Copy olmoearth_config.yaml to scratch space
        olmoearth_config_path = self.project_path / OLMOEARTH_CONFIG_FILE_NAME
        if not olmoearth_config_path.exists():
            raise FileNotFoundError(
                f"Required project file not found: {olmoearth_config_path}. "
                + "Expected unified OlmoEarthConfig format (olmoearth_config.yaml)."
            )

        scratch_config_path = Path(self.scratch_space.olmoearth_config_path)
        scratch_config_path.parent.mkdir(parents=True, exist_ok=True)
        shutil.copy2(olmoearth_config_path, scratch_config_path)

        # Create dataset directory
        Path(self.model_stage_paths.default_dataset_path).parent.mkdir(
            parents=True, exist_ok=True
        )

        # Symlink annotation feature files (still required for labeled data prep)
        annotation_files: list[tuple[Path, Path]] = [
            (
                self.project_path / ANNOTATION_FEATURES_FILE_NAME,
                Path(self.model_stage_paths.annotation_features_path),
            ),
            (
                self.project_path / ANNOTATION_TASK_FEATURES_FILE_NAME,
                Path(self.model_stage_paths.annotation_task_features_path),
            ),
        ]

        for target, link in annotation_files:
            if not target.exists():
                raise FileNotFoundError(f"Required project file not found: {target}")

            if not link.exists():
                link.parent.mkdir(parents=True, exist_ok=True)
                link.symlink_to(target)

    def prepare_labeled_windows(self) -> PrepareLabeledWindowsTaskResults:
        logger.info("Preparing labeled windows")
        task_args = PrepareLabeledWindowsTaskArgs(
            dataset_path=str(self.dataset_path),
            scratch_path=str(self.scratch_path),
        )
        return PrepareLabeledWindowsStepDefinition().run(task_args)

    def build_dataset_from_windows(self) -> DatasetBuildFromWindowsTaskResults:
        logger.info("Building dataset from windows")
        task_args = DatasetBuildFromWindowsTaskArgs(
            dataset_path=str(self.dataset_path),
            olmoearth_config_path=self.scratch_space.olmoearth_config_path,
        )
        return DatasetBuildFromWindowsStepDefinition().run(task_args)

    def fine_tune(self) -> FineTuneTaskResults:
        logger.info("Fine tuning")
        task_args = FineTuneTaskArgs(
            dataset_path=str(self.dataset_path),
            scratch_path=str(self.scratch_path),
            step_id=self.step_id,
        )
        return FineTuningStepDefinition().run(task_args)

    def run_pipeline(self) -> None:
        logger.info("Running full fine tune pipeline")
        self.prepare_labeled_windows()
        self.build_dataset_from_windows()
        self.fine_tune()



from enum import StrEnum
from typing import Annotated, Literal

from pydantic import BaseModel, Field
from typing_extensions import TypeAliasType

MAX_TIMESTEPS = 12  # OlmoEarth's max number of timesteps


class TemporalityName(StrEnum):
    SAMPLED = "sampled"
    REPEATING_INTERVAL = "repeating_interval"


class Month(StrEnum):
    JANUARY = "January"
    FEBRUARY = "February"
    MARCH = "March"
    APRIL = "April"
    MAY = "May"
    JUNE = "June"
    JULY = "July"
    AUGUST = "August"
    SEPTEMBER = "September"
    OCTOBER = "October"
    NOVEMBER = "November"
    DECEMBER = "December"


class SampledTemporality(BaseModel):
    """
    Represents a temporal configuration that samples data from a given duration.
    """
    name: Literal[TemporalityName.SAMPLED]
    duration_days: int = Field(ge=0, description="The length of time in which to sample data.")
    num_samples: int = Field(ge=1, le=MAX_TIMESTEPS, description="The number of samples to take or construct within the duration.")


class RepeatingIntervalTemporality(BaseModel):
    """
    Represents a temporal configuration to sample/generate a single image for each of a fixed number
    of periods of equal duration.
    """
    name: Literal[TemporalityName.REPEATING_INTERVAL]
    period_duration_days: int = Field(ge=0, description="The length of time a single period lasts for.")
    num_periods: int = Field(ge=1, le=MAX_TIMESTEPS, description="The number of periods total, each `period_duration_days` long.")
    allowed_start_month_constraints: set[Month] | None = Field(default=None, description="The months that are valid for starting the repeating interval in.")


# TODO: pre_post_sampled, pre_post_interval


Temporality = TypeAliasType(
    "Temporality",
    Annotated[
        SampledTemporality | RepeatingIntervalTemporality,
        Field(discriminator="name")
    ]
)



import logging
from typing import cast

from upath import UPath

from olmoearth_run.runner.steps.base_step_definition import BaseStepDefinition
from olmoearth_run.runner.tools.olmoearth_config.loaders.config import ConfigClassLoader
from olmoearth_run.runner.tools.olmoearth_config.loaders.prediction_tools import PredictionToolsClassLoader
from olmoearth_run.runner.tools.olmoearth_config.utils import ConfigUtils
from olmoearth_run.runner.tools.olmoearth_run_config_loader import OlmoEarthRunConfigLoader
from olmoearth_run.runner.tools.postprocessors.postprocess_interface import PostprocessInterfaceRaster, PostprocessInterfaceVector
from olmoearth_run.shared.models.api.task_args import CombinePartitionsTaskArgs
from olmoearth_run.shared.models.api.task_results import CombinePartitionsTaskResults, InferenceResultsDataType
from olmoearth_run.shared.models.model_stage_paths import ModelStagePaths
from olmoearth_run.shared.models.olmoearth_run_config import OlmoEarthRunConfig
from olmoearth_run.shared.models.prediction_scratch_space import PredictionScratchSpace
from olmoearth_shared.api.run.prediction_geometry import PredictionResultCollection
from olmoearth_shared.models.olmoearth_config.olmoearth_config import OlmoEarthConfig

logger = logging.getLogger(__name__)


class CombinePartitionsStepDefinition(BaseStepDefinition[CombinePartitionsTaskArgs, CombinePartitionsTaskResults]):
    scratch: PredictionScratchSpace
    config: OlmoEarthRunConfig | OlmoEarthConfig

    def run(self, task_args: CombinePartitionsTaskArgs) -> CombinePartitionsTaskResults:
        self.scratch = PredictionScratchSpace(root_path=task_args.scratch_path)
        model_stage_paths = ModelStagePaths(root_path=task_args.model_stage_root_path)
        UPath(model_stage_paths.root_path).mkdir(parents=True, exist_ok=True)

        # Load config - check for unified config first
        olmoearth_config_path = UPath(self.scratch.olmoearth_config_path)
        if olmoearth_config_path.exists():
            self.config = ConfigClassLoader.load_olmoearth_config(str(olmoearth_config_path))
            result_data_type = ConfigUtils.get_output_data_type(self.config)
        else:
            self.config = OlmoEarthRunConfigLoader.load_olmoearth_run_config(model_stage_paths.olmoearth_run_config_path)
            result_data_type = self.config.inference_results_config.data_type

        if result_data_type == InferenceResultsDataType.VECTOR:
            return self._combine_vector(task_args)
        return self._combine_raster(task_args)

    def _combine_raster(self, task_args: CombinePartitionsTaskArgs) -> CombinePartitionsTaskResults:
        logger.debug(f"Combining raster partitions: {task_args.partition_ids} args: {task_args}")
        postprocessor = cast(PostprocessInterfaceRaster, PredictionToolsClassLoader.load_request_postprocessor(self.config))

        # The scratch space keeps track of partition directories.
        # But partition processors can write one or multiple GeoTIFFs under that directory.
        # So we have to scan the contents across the directories.
        partition_result_paths: list[UPath] = []
        for partition_id in task_args.partition_ids:
            partition_result_dir = self.scratch.get_partition_result_raster_dir(partition_id)
            if not partition_result_dir.exists():
                continue
            partition_result_paths.extend(partition_result_dir.iterdir())

        prediction_result_raster_dir = UPath(self.scratch.prediction_result_raster_dir)
        prediction_result_raster_dir.mkdir(parents=True, exist_ok=True)
        saved_raster_paths = postprocessor.process_dataset(partition_result_paths, prediction_result_raster_dir)

        return CombinePartitionsTaskResults(generated_file_paths=[str(raster_path) for raster_path in saved_raster_paths])

    def _combine_vector(self, task_args: CombinePartitionsTaskArgs) -> CombinePartitionsTaskResults:
        logger.debug(f"Combining vector partitions: {task_args.partition_ids} args: {task_args}")
        postprocessor = cast(PostprocessInterfaceVector, PredictionToolsClassLoader.load_request_postprocessor(self.config))

        partition_results: list[PredictionResultCollection] = []
        partition_result_paths = [self.scratch.get_partition_result_vector_path(partition_id) for partition_id in task_args.partition_ids]
        for partition_result_path in partition_result_paths:
            partition_result_json = UPath(partition_result_path).read_text()
            partition_result = PredictionResultCollection.model_validate_json(partition_result_json)
            partition_results.append(partition_result)

        combined_result = postprocessor.process_dataset(partition_results)
        self.scratch.write_prediction_result_vector(combined_result)
        return CombinePartitionsTaskResults(generated_file_paths=[self.scratch.prediction_result_vector_path])



from prometheus_client import Counter, Histogram
from prometheus_client.utils import INF

_DATASET_BUILDER_DURATION_HISTOGRAM_BUCKETS = (
    5.0,
    10.0,
    15.0,
    20.0,
    30.0,
    40.0,
    50.0,
    60.0,  # 1 minute
    75.0,  # 1m 15s
    90.0,  # 1m 30s
    105.0,  # 1m 45s
    120.0,  # 2 minutes
    180.0,  # 3 minutes
    240.0,  # 4 minutes
    300.0,  # 5 minutes
    360.0,  # 6 minutes
    420.0,  # 7 minutes
    480.0,  # 8 minutes
    540.0,  # 9 minutes
    600.0,  # 10 minutes
    900.0,  # 15 minutes
    1200.0,  # 20 minutes
    1500.0,  # 25 minutes
    1800.0,  # 30 minutes
    2100.0,  # 35 minutes
    2400.0,  # 40 minutes
    2700.0,  # 45 minutes
    3000.0,  # 50 minutes
    3600.0,  # 1 hour
    4200.0,  # 1h 10m
    4800.0,  # 1h 20m
    5400.0,  # 1h 30m
    6000.0,  # 1h 40m
    6600.0,  # 1h 50m
    7200.0,  # 2 hours
    9000.0,  # 2h 30m
    10800.0,  # 3 hours
    12600.0,  # 3h 30m
    14400.0,  # 4 hours
    16200.0,  # 4h 30m
    18000.0,  # 5 hours
    19800.0,  # 5h 30m
    21600.0,  # 6 hours
    23400.0,  # 6h 30m
    25200.0,  # 7 hours
    27000.0,  # 7h 30m
    28800.0,  # 8 hours
    INF,
)


# Private Prometheus metrics
_load_windows_loaded_counter = Counter(
    'runner_dataset_builder_loading_windows_loaded_total',
    'Number of windows loaded',
)

_prepare_windows_processed_counter = Counter(
    'runner_dataset_builder_prepare_windows_processed_total',
    'Number of windows processed for preparation',
)

_prepare_window_layers_handled_counter = Counter(
    'runner_dataset_builder_prepare_layers_windows_handled_total',
    'Number of window layers handled per data source, with skipped status',
    labelnames=['data_source', 'skipped'],
)

_prepare_window_layers_handle_attempts_counter = Counter(
    'runner_dataset_builder_prepare_layers_windows_handle_attempts_total',
    'Number of attempts to handle window layers per data source',
    labelnames=['data_source'],
)

_prepare_window_rejection_ratio_error_counter = Counter(
    'runner_dataset_builder_prepare_rejection_ratio_error_total',
    'Number of times window success rate fell below threshold',
)

_ingest_jobs_processed_counter = Counter(
    'runner_dataset_builder_ingest_jobs_processed_total',
    'Number of jobs processed for ingestion',
)

_ingest_geometries_ingested_counter = Counter(
    'runner_dataset_builder_ingest_geometries_ingested_total',
    'Number of geometries ingested, by data source, with outcome',
    labelnames=['data_source', 'outcome'],
)

_materialize_windows_processed_counter = Counter(
    'runner_dataset_builder_materialize_windows_processed_total',
    'Number of windows processed for materialization',
)

_materialize_window_layers_materialized_counter = Counter(
    'runner_dataset_builder_materialize_layers_windows_materialized_total',
    'Number of window layers materialized, by data source',
    labelnames=['data_source'],
)

_materialize_window_layers_materialize_attempts_counter = Counter(
    'runner_dataset_builder_materialize_layers_windows_materialize_attempts_total',
    'Number of attempts to materialize window layers, by data source',
    labelnames=['data_source'],
)

# Duration histograms
_prepare_layer_duration_histogram = Histogram(
    'runner_dataset_builder_prepare_layer_duration_seconds',
    'Duration in seconds per layer for prepare operations',
    labelnames=['data_source'],
    buckets=_DATASET_BUILDER_DURATION_HISTOGRAM_BUCKETS,
)

_prepare_duration_histogram = Histogram(
    'runner_dataset_builder_prepare_duration_seconds',
    'Duration in seconds for prepare operations',
    buckets=_DATASET_BUILDER_DURATION_HISTOGRAM_BUCKETS,
)

_ingest_layer_duration_histogram = Histogram(
    'runner_dataset_builder_ingest_layer_duration_seconds',
    'Duration in seconds per layer for ingest operations',
    labelnames=['data_source'],
    buckets=_DATASET_BUILDER_DURATION_HISTOGRAM_BUCKETS,
)

_ingest_duration_histogram = Histogram(
    'runner_dataset_builder_ingest_duration_seconds',
    'Duration in seconds for ingest operations',
    buckets=_DATASET_BUILDER_DURATION_HISTOGRAM_BUCKETS,
)

_materialize_layer_duration_histogram = Histogram(
    'runner_dataset_builder_materialize_layer_duration_seconds',
    'Duration in seconds per layer for materialize operations',
    labelnames=['data_source'],
    buckets=_DATASET_BUILDER_DURATION_HISTOGRAM_BUCKETS,
)

_materialize_duration_histogram = Histogram(
    'runner_dataset_builder_materialize_duration_seconds',
    'Duration in seconds for materialize operations',
    buckets=_DATASET_BUILDER_DURATION_HISTOGRAM_BUCKETS,
)


class DatasetBuilderMetrics:
    """Metrics for dataset building operations."""

    @staticmethod
    def record_windows_loaded(count: int) -> None:
        """Record windows loaded."""
        _load_windows_loaded_counter.inc(count)

    @staticmethod
    def record_windows_processed(count: int) -> None:
        """Record windows processed for preparation."""
        _prepare_windows_processed_counter.inc(count)

    @staticmethod
    def record_window_layers_handled(data_source: str, skipped: bool, count: int) -> None:
        """Record window layers handled per data source."""
        _prepare_window_layers_handled_counter.labels(
            data_source=data_source,
            skipped=str(skipped)
        ).inc(count)

    @staticmethod
    def record_window_layers_handle_attempts(data_source: str, count: int) -> None:
        """Record attempts to handle window layers."""
        _prepare_window_layers_handle_attempts_counter.labels(
            data_source=data_source
        ).inc(count)

    @staticmethod
    def record_window_rejection_ratio_error(count: int) -> None:
        """Record when window success rate falls below threshold."""
        _prepare_window_rejection_ratio_error_counter.inc(count)

    @staticmethod
    def record_ingest_jobs_processed(count: int) -> None:
        """Record ingestion jobs processed."""
        _ingest_jobs_processed_counter.inc(count)

    @staticmethod
    def record_geometries_ingested(data_source: str, outcome: str, count: int) -> None:
        """Record geometries ingested."""
        _ingest_geometries_ingested_counter.labels(
            data_source=data_source,
            outcome=outcome
        ).inc(count)

    @staticmethod
    def record_windows_materialized(count: int) -> None:
        """Record windows processed for materialization."""
        _materialize_windows_processed_counter.inc(count)

    @staticmethod
    def record_window_layers_materialized(data_source: str, count: int) -> None:
        """Record window layers materialized."""
        _materialize_window_layers_materialized_counter.labels(
            data_source=data_source
        ).inc(count)

    @staticmethod
    def record_window_layers_materialize_attempts(data_source: str, count: int) -> None:
        """Record attempts to materialize window layers."""
        _materialize_window_layers_materialize_attempts_counter.labels(
            data_source=data_source
        ).inc(count)

    @staticmethod
    def record_prepare_layer_duration(data_source: str, duration: float) -> None:
        """Record per-layer prepare operation duration."""
        _prepare_layer_duration_histogram.labels(data_source=data_source).observe(duration)

    @staticmethod
    def record_prepare_duration(duration: float) -> None:
        """Record overall prepare operation duration."""
        _prepare_duration_histogram.observe(duration)

    @staticmethod
    def record_ingest_layer_duration(data_source: str, duration: float) -> None:
        """Record per-layer ingest operation duration."""
        _ingest_layer_duration_histogram.labels(data_source=data_source).observe(duration)

    @staticmethod
    def record_ingest_duration(duration: float) -> None:
        """Record overall ingest operation duration."""
        _ingest_duration_histogram.observe(duration)

    @staticmethod
    def record_materialize_layer_duration(data_source: str, duration: float) -> None:
        """Record per-layer materialize operation duration."""
        _materialize_layer_duration_histogram.labels(data_source=data_source).observe(duration)

    @staticmethod
    def record_materialize_duration(duration: float) -> None:
        """Record overall materialize operation duration."""
        _materialize_duration_histogram.observe(duration)



import json
import logging

from rslearn.arg_parser import RslearnArgumentParser
from rslearn.lightning_cli import RslearnLightningCLI
from rslearn.train.data_module import RslearnDataModule
from rslearn.train.lightning_module import RslearnLightningModule
from upath import UPath

from olmoearth_run.config import OlmoEarthSettings
from olmoearth_run.runner.models.operational_context import ModelOperationalContext
from olmoearth_run.runner.steps.base_step_definition import BaseStepDefinition
from olmoearth_run.runner.tools.olmoearth_config.loaders.config import ConfigClassLoader
from olmoearth_run.runner.tools.olmoearth_config.transpilers.olmoearth_to_rslearn import transpile_and_write_model_config
from olmoearth_run.runner.tools.olmoearth_config.utils import ConfigUtils
from olmoearth_run.runner.tools.olmoearth_run_config_loader import OlmoEarthRunConfigLoader
from olmoearth_run.shared.models.api.step_type import StepType
from olmoearth_run.shared.models.model_stage_paths import ModelStagePaths
from olmoearth_run.shared.models.prediction_scratch_space import (
    PredictionScratchSpace,
    LEGACY_WINDOW_OUTPUT_LAYER_NAME,
)
from olmoearth_run.shared.models.rslearn_template_vars import RslearnTemplateVars
from olmoearth_run.shared.models.api.task_args import RunInferenceTaskArgs
from olmoearth_run.shared.models.api.task_results import (
    RunInferenceTaskResults,
)


logger = logging.getLogger(__name__)


class RunInferenceStepDefinition(BaseStepDefinition[RunInferenceTaskArgs, RunInferenceTaskResults]):
    def run(self, task_args: RunInferenceTaskArgs) -> RunInferenceTaskResults:
        model_stage_paths = ModelStagePaths(root_path=task_args.model_stage_root_path)
        dataset_path = UPath(task_args.dataset_path)
        scratch = PredictionScratchSpace(root_path=task_args.scratch_path)

        # Check if unified config exists - determines which code path to use
        olmoearth_config_path = UPath(scratch.olmoearth_config_path)
        if olmoearth_config_path.exists():
            return self._run_with_unified_config(task_args, scratch, dataset_path, model_stage_paths)
        else:
            return self._run_with_legacy_config(task_args, scratch, dataset_path, model_stage_paths)

    def _run_with_unified_config(
        self,
        task_args: RunInferenceTaskArgs,
        scratch: PredictionScratchSpace,
        dataset_path: UPath,
        model_stage_paths: ModelStagePaths,
    ) -> RunInferenceTaskResults:
        """Run inference using unified OlmoEarthConfig with JIT transpilation per partition."""
        logger.info(f"Using unified config path: {scratch.olmoearth_config_path}")
        olmoearth_config = ConfigClassLoader.load_olmoearth_config(scratch.olmoearth_config_path)

        for partition_id in task_args.partition_ids:
            logger.info(f"Running inference step for partition {partition_id}")
            rslearn_groups = [scratch.get_group(partition_id)]

            # Build ops context with partition-specific groups
            ops_context = ModelOperationalContext(
                step_type=StepType.RUN_INFERENCE,
                dataset_path=str(dataset_path),
                num_data_worker_processes=OlmoEarthSettings.NUM_WORKERS,
                predict_groups=rslearn_groups,
                model_stage_paths=model_stage_paths,
            )

            # JIT transpile to partition-specific path
            model_config_path = UPath(scratch.get_partition_model_config_path(partition_id))
            transpile_and_write_model_config(olmoearth_config, ops_context, model_config_path)

            predict_args = [
                "predict",
                "--config", str(model_config_path),
                "--ckpt_path", model_stage_paths.checkpoint_path,
            ]
            logger.info(f"Running inference on groups {rslearn_groups} with args: {predict_args}")
            self._run_inference_for_groups(predict_args)

        # Determine data type from unified config output schema
        data_type = ConfigUtils.get_output_data_type(olmoearth_config)
        return RunInferenceTaskResults(inference_results_data_type=data_type)

    def _run_with_legacy_config(
        self,
        task_args: RunInferenceTaskArgs,
        scratch: PredictionScratchSpace,
        dataset_path: UPath,
        model_stage_paths: ModelStagePaths,
    ) -> RunInferenceTaskResults:
        """
        Run inference using legacy OlmoEarthRunConfig with env vars and CLI overrides.

        TODO chrisw: delete this code path once we've fully migrated to the unified config.
        """
        logger.info("Using legacy config path")
        olmoearth_run_config = OlmoEarthRunConfigLoader.load_olmoearth_run_config(model_stage_paths.olmoearth_run_config_path)

        rslearn_env_vars = RslearnTemplateVars(
            DATASET_PATH=str(dataset_path),
            EXTRA_FILES_PATH=str(model_stage_paths.extra_model_files_path),
            TRAINER_DATA_PATH=str(model_stage_paths.trainer_checkpoints_path),
            NUM_WORKERS=OlmoEarthSettings.NUM_WORKERS,
            PREDICTION_OUTPUT_LAYER=LEGACY_WINDOW_OUTPUT_LAYER_NAME,
            WANDB_PROJECT=None,  # Inference doesn't need wandb logging
            WANDB_NAME=None,     # Inference doesn't need wandb logging
            WANDB_ENTITY=None,   # Inference doesn't need wandb logging
        )

        # TODO: how much of the rest of this can go in RslearnTemplateVars?
        for partition_id in task_args.partition_ids:
            logger.info(f"Running inference step for partition {partition_id}")
            rslearn_groups = [scratch.get_group(partition_id)]
            predict_args = [
                "predict",
                "--config", model_stage_paths.model_config_path,
                "--ckpt_path", model_stage_paths.checkpoint_path,
                # only run inference on the group
                "--data.predict_config.groups", json.dumps(rslearn_groups),
                "--trainer.logger", "false",  # disable wandb logging
                "--data.init_args.predict_config.output_layer", LEGACY_WINDOW_OUTPUT_LAYER_NAME,
            ]
            logger.debug(f"Setting rslearn env vars: {rslearn_env_vars}")
            logger.info(f"Running inference on groups {rslearn_groups} with args: {predict_args}")
            with rslearn_env_vars.temp_env():
                self._run_inference_for_groups(predict_args)

        return RunInferenceTaskResults(inference_results_data_type=olmoearth_run_config.inference_results_config.data_type)

    def _run_inference_for_groups(self, predict_args: list[str]) -> None:
        # We need to catch SystemExit because rslearn or any of its dependencies might call sys.exit() which will
        # wrest control away from us and crash the step without informing the orchestrator.
        try:
            RslearnLightningCLI(
                model_class=RslearnLightningModule,
                datamodule_class=RslearnDataModule,
                parser_class=RslearnArgumentParser,
                args=predict_args,
                subclass_mode_model=True,
                subclass_mode_data=True,
                save_config_kwargs={"overwrite": True},
            )
        except SystemExit as e:
            if e.code != 0:
                raise RuntimeError(f"rslearn exited with code {e.code}") from e



import yaml
from pydantic import BaseModel, Field

from olmoearth_shared.models.olmoearth_config.data.data import Data
from olmoearth_shared.models.olmoearth_config.input_preprocessing.input_preprocessing import InputPreprocessing
from olmoearth_shared.models.olmoearth_config.labeled_data_prep.labeled_data_prep import LabeledDataPrep
from olmoearth_shared.models.olmoearth_config.model.model import Model
from olmoearth_shared.models.olmoearth_config.prediction_requests.prediction_requests import PredictionRequests
from olmoearth_shared.models.olmoearth_config.training.training import Training


class OlmoEarthConfig(BaseModel):
    """
    Unified configuration for OlmoEarth: data requirements, model, dataset prep, training, prediction.
    """

    config_version: str = Field(description="The version of the configuration")
    data: Data = Field(description="Specifies the required data modalities, temporality of the data to retrieve, and the output schema (when a `model` is defined).")
    model: Model | None = Field(default=None, description="Defines the model architecture and the tasks to perform during inference.")
    input_preprocessing: InputPreprocessing | None = Field(default=None, description="Defines how input tensors are prepared or transformed before being provided to the model.")
    labeled_data_prep: LabeledDataPrep | None = Field(default=None, description="Controls the preparation of labeled data for training.")
    training: Training | None = Field(default=None, description="Configuration for the training")
    prediction_requests: PredictionRequests | None = Field(default=None, description="Controls the prediction requests for the model.")

    @property
    def version_major(self) -> int:
        """Extract major version component from config_version (e.g., '1.2.3' -> 1)."""
        return self._parse_semver_part(0)

    @property
    def version_minor(self) -> int:
        """Extract minor version component from config_version (e.g., '1.2.3' -> 2)."""
        return self._parse_semver_part(1)

    @property
    def version_micro(self) -> int:
        """Extract micro/patch version component from config_version (e.g., '1.2.3' -> 3)."""
        return self._parse_semver_part(2)

    def _parse_semver_part(self, index: int) -> int:
        """Parse a specific part of the semver string, handling pre-release/build metadata."""
        # Strip any pre-release (-beta.1) or build metadata (+build.123)
        base_version = self.config_version.split('-')[0].split('+')[0]
        parts = base_version.split('.')
        return int(parts[index]) if len(parts) > index else 0

    @classmethod
    def from_yaml(cls, yaml_content: str) -> "OlmoEarthConfig":
        """
        Parse OlmoEarth configuration from a YAML string.

        Raises:
            ValueError: If the YAML content is malformed or doesn't match the expected structure
        """
        try:
            raw_config = yaml.safe_load(yaml_content)
        except yaml.YAMLError as e:
            raise ValueError("Invalid YAML in olmoearth_run_config") from e

        try:
            return cls.model_validate(raw_config)
        except Exception as e:  # Catch ValidationError and other pydantic errors
            raise ValueError(f"Invalid olmoearth_config structure. {e}") from e



from pydantic import BaseModel

from olmoearth_shared.models.olmoearth_config.data.modalities import Modalities
from olmoearth_shared.models.olmoearth_config.data.output import Output
from olmoearth_shared.models.olmoearth_config.data.temporality import Temporality


class Data(BaseModel):
    temporality: Temporality
    modalities: Modalities
    output: Output | None



import logging
import re
import tempfile
from collections import defaultdict
from contextlib import ExitStack, contextmanager
from dataclasses import dataclass
from pathlib import Path
from typing import Iterator, Literal

import numpy as np
import rasterio  # type: ignore[import-untyped]
from rasterio.merge import merge  # type: ignore[import-untyped]
from rasterio.session import GSSession, DummySession  # type: ignore[import-untyped]
from shapely.geometry import Polygon, box
from upath import UPath

from olmoearth_run.runner.tools.mask_utils import MaskConfig, apply_mask_to_raster_data, \
    get_first_intersecting_mask_for_raster
from olmoearth_run.runner.tools.postprocessors.grid_utils import group_polygons_by_grid
from olmoearth_run.runner.tools.postprocessors.postprocess_interface import PostprocessInterfaceRaster
from olmoearth_shared.api.run.prediction_geometry import PredictionRequestFeature
from olmoearth_shared.tools.gcs_tools import is_gcs_path, write_file_to_gcs, download_gcs_files

logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class LinearValueTransform:
    """Linear transform: output = scale * input + offset.

    Uses discriminated union pattern to support other transform types in the future.
    """
    type: Literal["linear"] = "linear"
    scale: float = 1.0
    offset: float = 0.0
    output_dtype: Literal["uint8", "int16", "int32", "float32"] | None = None

    def apply(self, data: np.ndarray) -> np.ndarray:
        result = data * self.scale + self.offset
        if self.output_dtype:
            dtype = np.dtype(self.output_dtype)
            if np.issubdtype(dtype, np.integer):
                info = np.iinfo(dtype)
                result = np.clip(result, info.min, info.max)
            result = result.astype(dtype)
        return result


# Union type for future extensibility (e.g., LogValueTransform, etc.)
ValueTransform = LinearValueTransform


@dataclass
class RasterMetadata:
    """Cached metadata for a GeoTIFF file."""
    path: UPath
    crs_string: str
    geometry: Polygon  # Shapely polygon representing raster bounds in CRS coordinate space
    pixel_size_x: float  # Resolution of x dimension. (1 pixel = K units in CRS). This is meters per pixel for UTM projections.
    pixel_size_y: float  # resolution of y dimension


class CombineGeotiff(PostprocessInterfaceRaster):
    def __init__(
        self,
        max_pixels_per_dimension: int = 10000,
        nodata_value: int | float = 0,
        value_transform: ValueTransform | None = None,
    ):
        self.max_pixels_per_dimension = max_pixels_per_dimension
        self.nodata_value = nodata_value
        self.value_transform = value_transform
        self._mask_config: MaskConfig | None = None

    def set_mask_config(self, mask_config: MaskConfig | None) -> None:
        """Configure masking to apply after merging.
        This is a bit of a hack because we really want to run masking when we already have the geotiffs open in memory.
        And so we run the masking during this combine step.   In reality basically all raster flows use this CombineGeotiff
        postprocessor; we might want to consider making this the default behavior and providing some other way to enable
        custom postprocessing.
        """
        self._mask_config = mask_config

    def process_window(self, window_request: PredictionRequestFeature, window_result_path: UPath) -> None:
        raise NotImplementedError("CombineGeotiff does not support combining windows")

    def process_partition(self, partition_window_results: list[UPath], output_dir: UPath) -> list[UPath]:
        if not partition_window_results:
            return []
        return self._combine_geotiffs(partition_window_results, output_dir)

    def process_dataset(self, all_partitions_result_paths: list[UPath], output_dir: UPath) -> list[UPath]:
        if not all_partitions_result_paths:
            return []
        return self._combine_geotiffs(all_partitions_result_paths, output_dir)

    def _combine_geotiffs(self, input_paths: list[UPath], output_directory: UPath) -> list[UPath]:
        """
        Combine multiple GeoTIFF files, grouping by CRS and using grid-based tiling
        to avoid creating extremely large output files.
        """
        logger.info(f"Combining {len(input_paths)} rasters")

        if not input_paths:
            logger.warning("No input paths provided to CombineGeotiff")
            return []

        # Group input files by CRS. Since we cannot mix CRS in output files.
        all_metadata = [self._get_raster_metadata(p) for p in input_paths]
        crs_groups = self._group_by_crs(all_metadata)
        logger.info(f"Grouped {len(input_paths)} rasters into {len(crs_groups)} CRS groups: {list(crs_groups.keys())}")

        output_files = []

        # Process each CRS group independently
        for crs_string, metadata_list in crs_groups.items():
            # Sanitize CRS string for filename (e.g., EPSG:32610 -> epsg32610)
            crs_safe = re.sub(r'[^a-zA-Z0-9]', '', crs_string)
            logger.info(f"Processing CRS group {crs_string} with {len(metadata_list)} rasters")

            # Calculate max_size in coordinate units (use first raster's pixel size as reference)
            max_size = self.max_pixels_per_dimension * metadata_list[0].pixel_size_x

            # Group to ensure that we don't end up with any single output file over max_size
            grid_groups = group_polygons_by_grid(
                metadata_list,
                max_size,
                get_geometry=lambda m: m.geometry
            )

            logger.info(f"Assigned {len(metadata_list)} rasters to {len(grid_groups)} grid cells")

            # Merge each grid group into a single output raster. Downloading the files for each cell.
            for idx, ((grid_x, grid_y), metadata_items) in enumerate(grid_groups.items()):
                # Create output filename with CRS and grid coordinates
                output_filename = f"result_{crs_safe}_{idx}.tif"

                paths = [m.path for m in metadata_items]
                logger.info(f"Merging grid cell ({grid_x}, {grid_y}) with {len(paths)} rasters -> {output_filename}")

                # Merge the rasters in this grid cell
                with managed_raster_paths(paths, output_directory) as (local_paths, local_output_dir):
                    local_output = local_output_dir / output_filename
                    self._merge_grid_cell(local_paths, local_output)
                    output_files.extend(convert_outputs([local_output], output_directory))

        logger.info(f"Combined {len(input_paths)} rasters into {len(output_files)} output files")
        return output_files

    @staticmethod
    def _get_raster_metadata(input_path: UPath) -> RasterMetadata:
        """Loads the bounds/CRS for a GeoTIFF file, reading headers-only directly from GCS if needed."""
        session = GSSession() if is_gcs_path(str(input_path)) else DummySession()
        with rasterio.Env(session=session), rasterio.open(str(input_path)) as src:
            bounds = src.bounds
            return RasterMetadata(
                path=input_path,
                crs_string=src.crs.to_string().lower() if src.crs else "unknown",
                geometry=box(bounds.left, bounds.bottom, bounds.right, bounds.top),
                pixel_size_x=abs(src.transform[0]),
                pixel_size_y=abs(src.transform[4])
            )

    @staticmethod
    def _group_by_crs(metadata_list: list[RasterMetadata]) -> dict[str, list[RasterMetadata]]:
        """
        Returns dictionary mapping CRS string (e.g., 'EPSG:32610') to list of metadata objects
        """
        crs_groups: dict[str, list[RasterMetadata]] = defaultdict(list)
        for metadata in metadata_list:
            crs_groups[metadata.crs_string].append(metadata)
        return dict(crs_groups)

    def _merge_grid_cell(self, paths: list[Path], output_path: Path) -> None:
        """Merge a list of rasters into a single output GeoTIFF."""

        if len(paths) == 1:
            # Just one file, copy it and set nodata value
            logger.info(f"Only one raster in grid cell, copying directly: {paths[0]}")
            with rasterio.open(paths[0]) as src:
                data = src.read()
                if self._mask_config:
                    mask_file = get_first_intersecting_mask_for_raster(self._mask_config.mask_files, src.bounds, src.crs)
                    apply_mask_to_raster_data(data, src.transform, src.crs, mask_file, self._mask_config.valid_values, self.nodata_value)
                if self.value_transform:
                    data = self.value_transform.apply(data)
                meta = src.meta.copy()
                meta.update({
                    "nodata": self.nodata_value,
                    "compress": "lzw",
                    "dtype": data.dtype,
                })
                with rasterio.open(output_path, "w", **meta) as dst:
                    dst.write(data)
            return

        # Open all input rasters with proper cleanup
        with ExitStack() as stack:
            input_rasters = [stack.enter_context(rasterio.open(path)) for path in paths]

            merged_array, transform = merge(input_rasters, nodata=self.nodata_value)
            # Apply mask in-memory if configured
            if self._mask_config:
                merged_bounds = rasterio.transform.array_bounds(merged_array.shape[1], merged_array.shape[2], transform)
                mask_file = get_first_intersecting_mask_for_raster(self._mask_config.mask_files, merged_bounds, input_rasters[0].crs)
                apply_mask_to_raster_data(merged_array, transform, input_rasters[0].crs, mask_file, self._mask_config.valid_values, self.nodata_value)
            if self.value_transform:
                merged_array = self.value_transform.apply(merged_array)

            meta = input_rasters[0].meta.copy()
            meta.update({
                "driver": "GTiff",
                "height": merged_array.shape[1],
                "width": merged_array.shape[2],
                "transform": transform,
                "compress": "lzw",
                "nodata": self.nodata_value,
                "dtype": merged_array.dtype,
            })

            with rasterio.open(output_path, "w", **meta) as destination:
                destination.write(merged_array)

        logger.debug(f"Merged {len(paths)} rasters to {output_path}")


@contextmanager
def managed_raster_paths(input_paths: list[UPath], output_directory: UPath) -> Iterator[tuple[list[Path], Path]]:
    """Provides an abstraction to ensure input/output paths are local paths."""
    with tempfile.TemporaryDirectory() as temp_dir:
        local_inputs = convert_inputs_to_local(input_paths, temp_dir)
        if is_gcs_path(str(output_directory)):
            # If the ultimate output dir is GCS, then create a temp output dir. Results will get copied later.
            local_output_dir = Path(temp_dir) / "outputs"
            local_output_dir.mkdir(exist_ok=True)
            yield local_inputs, local_output_dir
        else:
            # If the ultimate output dir is local, we can write straight there!
            yield local_inputs, Path(str(output_directory))


def convert_inputs_to_local(input_files: list[UPath], temp_dir: str) -> list[Path]:
    """Copies GCS files to local temp dir, or just skips if the files are already local. """
    is_gcs = is_gcs_path(str(input_files[0]))
    if not is_gcs:
        # Local paths, pass through
        return [Path(str(p)) for p in input_files]

    temp_path = Path(temp_dir)
    temp_inputs_dir = temp_path / "inputs"
    temp_inputs_dir.mkdir()

    # Copy each input file to temp directory
    logger.info(f"Copying {len(input_files)} input rasters from GCS to local temp directory")
    local_input_paths = download_gcs_files([str(f) for f in input_files], temp_inputs_dir)
    return local_input_paths


def convert_outputs(local_output_files: list[Path], output_dir: UPath) -> list[UPath]:
    """Copies local output files to GCS if needed, or just returns local paths."""
    is_gcs = is_gcs_path(str(output_dir))
    if not is_gcs:
        return [UPath(f) for f in local_output_files]

    # Copy outputs back to GCS
    output_files = []
    for local_file in local_output_files:
        logger.info(f"Copying {local_file} to GCS {output_dir / local_file.name}...")
        write_file_to_gcs(str(local_file), str(output_dir / local_file.name))
        output_files.append(UPath(output_dir / local_file.name))
    return output_files



import random

from olmoearth_run.runner.tools.data_splitters.data_splitter_interface import (
    DataSplitterInterface,
)
from olmoearth_run.shared.models.data_split_type import DataSplitType
from olmoearth_run.runner.models.training.labeled_data import LabeledWindow
from olmoearth_shared.models.olmoearth_config.labeled_data_prep.data_splitter import SplitProportions


class RandomDataSplitter(DataSplitterInterface):
    """Data splitter that randomly assigns splits based on specified proportions."""

    def __init__(
        self,
        train_prop: float,
        val_prop: float,
        test_prop: float,
        seed: int | None = None,
    ):
        """
        Initialize random data splitter with proportions for each split.

        Args:
            train_prop: Proportion of data for training (0.0 to 1.0)
            val_prop: Proportion of data for validation (0.0 to 1.0)
            test_prop: Proportion of data for testing (0.0 to 1.0)
            seed: Optional random seed for reproducible splits

        Raises:
            ValueError: If proportions are negative or don't sum to 1.0
        """
        self.split_proportions = SplitProportions(train_prop=train_prop, val_prop=val_prop, test_prop=test_prop)

        # Set up random state
        self.random = random.Random(seed)

    def choose_split_for_window(self, labeled_window: LabeledWindow) -> DataSplitType:
        """
        Randomly choose a data split based on configured proportions.

        This implementation ignores all window context properties and makes a purely
        random assignment based on the proportions specified at initialization.

        Args:
            window_context: Ignored (all properties ignored for random assignment)

        Returns:
            Randomly assigned data split
        """
        # Generate random number and assign split based on cumulative proportions
        rand_val = self.random.random()

        if rand_val < self.split_proportions.train_prop:
            return DataSplitType.TRAIN
        elif (
            rand_val
            < self.split_proportions.train_prop + self.split_proportions.val_prop
        ):
            return DataSplitType.VAL
        else:
            return DataSplitType.TEST



from enum import StrEnum


class StepType(StrEnum):
    """
    An enumerated type of all known steps in the OlmoEarthRun system.
    Different Workflows might implement these steps in different ways.
    """
    FINE_TUNE = "FINE_TUNE"
    PREPARE_LABELED_WINDOWS = "PREPARE_LABELED_WINDOWS"
    CREATE_PARTITIONS = "CREATE_PARTITIONS"
    DATASET_BUILD = "DATASET_BUILD"
    DATASET_BUILD_FROM_WINDOWS = "DATASET_BUILD_FROM_WINDOWS"
    RUN_INFERENCE = "RUN_INFERENCE"
    POSTPROCESS_PARTITION = "POSTPROCESS_PARTITION"
    COMBINE_PARTITIONS = "COMBINE_PARTITIONS"
    MODEL_EVALUATION = "MODEL_EVALUATION"



import logging
import sys
import uuid
from os import PathLike
from pathlib import Path

from olmoearth_run.runner.steps.combine_partitions_step_definition import CombinePartitionsStepDefinition
from olmoearth_run.runner.steps.create_partitions_step_definition import CreatePartitionsStepDefinition
from olmoearth_run.runner.steps.dataset_build_step_definition import DatasetBuildStepDefinition
from olmoearth_run.runner.steps.postprocess_partition_step_definition import PostprocessPartitionStepDefinition
from olmoearth_run.runner.steps.run_inference_step_definition import RunInferenceStepDefinition
from olmoearth_run.runner.tools.olmoearth_run_config_loader import OlmoEarthRunConfigLoader
from olmoearth_run.shared.models.model_stage_paths import (
    CHECKPOINT_FILE_NAME,
    DATASET_CONFIG_FILE_NAME,
    MODEL_CONFIG_FILE_NAME,
    OLMOEARTH_RUN_CONFIG_FILE_NAME,
    ModelStagePaths,
)
from olmoearth_run.shared.models.prediction_scratch_space import PREDICTION_REQUEST_GEOMETRY_FILE_NAME, PredictionScratchSpace
from olmoearth_run.shared.models.rslearn_template_vars import RslearnTemplateVars
from olmoearth_run.shared.models.api.task_args import (
    CombinePartitionsTaskArgs,
    CreatePartitionsTaskArgs,
    DatasetBuildTaskArgs,
    PostprocessPartitionTaskArgs,
    RunInferenceTaskArgs,
)
from olmoearth_run.shared.models.api.task_results import (
    CombinePartitionsTaskResults,
    PostprocessPartitionTaskResults,
    RunInferenceTaskResults,
    DatasetBuildTaskResults,
    InferenceResultsDataType,
)

logger = logging.getLogger(__name__)

NAMESPACE_ES_PREDICT_RUNNER = uuid.UUID('955274c2-5831-4690-b1bd-e48f50fcbc5b')


class OlmoEarthRunPredictRunner:
    def __init__(self,
                 project_path: PathLike,
                 scratch_path: PathLike,
                 request_geometry_path: PathLike | None = None,
                 checkpoint_path: PathLike | None = None,
    ):
        self.project_path = Path(project_path)
        self.scratch_path = Path(scratch_path)
        self.dataset_path = f"{self.scratch_path}/dataset_0" #  /{DATASET_CONFIG_FILE_NAME}"

        self.stage_name = "model_stage_0"
        self.stage_path = self.scratch_path / self.stage_name
        self.project_name = self.project_path.name
        self.inference_results_data_type: InferenceResultsDataType | None = None

        self.scratch = PredictionScratchSpace(root_path=str(self.scratch_path))
        self.model_stage_paths = ModelStagePaths(root_path=str(self.stage_path))

        self.project_uuid_namespace = uuid.uuid5(NAMESPACE_ES_PREDICT_RUNNER, self.project_name)
        self.model_stage_id = uuid.uuid5(self.project_uuid_namespace, self.stage_name)

        self.checkpoint_path = Path(checkpoint_path or self.project_path / CHECKPOINT_FILE_NAME)
        self.request_geometry_path = Path(request_geometry_path or self.project_path / PREDICTION_REQUEST_GEOMETRY_FILE_NAME)

        self.olmoearth_run_config = OlmoEarthRunConfigLoader.load_olmoearth_run_config(str(self.project_path / OLMOEARTH_RUN_CONFIG_FILE_NAME))

        logger.info(f"Project Path: {self.project_path}")
        logger.info(f"Scratch Path: {self.scratch_path}")
        logger.info(f"Dataset Path: {self.dataset_path}")

        self._setup_project_env()

    def _setup_project_env(self) -> None:
        """Set up the project environment in the scratch space by creating necessary directories and symlinks."""
        Path(self.scratch_path).mkdir(parents=True, exist_ok=True)
        Path(self.dataset_path).mkdir(exist_ok=True)

        project_files : list[tuple[Path, Path]] = [
            (self.request_geometry_path, Path(self.scratch.prediction_request_geometry_path)),
            (self.project_path / DATASET_CONFIG_FILE_NAME, Path(self.dataset_path ) / "config.json"),
            (self.project_path / MODEL_CONFIG_FILE_NAME, Path(self.model_stage_paths.model_config_path)),
            (self.project_path / OLMOEARTH_RUN_CONFIG_FILE_NAME, Path(self.model_stage_paths.olmoearth_run_config_path)),
            (self.checkpoint_path, Path(self.model_stage_paths.checkpoint_path)),
        ]

        for (target, link) in project_files:
            if not link.exists():
                if not target.exists():
                    if target == self.checkpoint_path:
                        logger.warning(f"Checkpoint file {target} does not exist. Skipping checkpoint linking.")
                        continue
                    else:
                        logger.error(f"Required project file {target} does not exist.")
                        sys.exit(1)
                logger.info(f"{link} does not exist yet. Linking {target} to {link}")
                link.parent.mkdir(parents=True, exist_ok=True)
                link.symlink_to(target)
            else:
                logger.info(f"{link} already exists. Skipping linking {target} to {link}")

        with open(self.model_stage_paths.model_config_path, 'r') as f:
            model_config_yaml = f.read()
            RslearnTemplateVars.validate_yaml(model_config_yaml)

    def partition(self) -> list[str]:
        """Partition the original request feature into smaller features for processing"""
        partition_args = CreatePartitionsTaskArgs(
            scratch_path=str(self.scratch_path),
            model_stage_root_path=self.model_stage_paths.root_path,
            model_stage_id=self.model_stage_id,
            dataset_path=self.dataset_path,
        )
        return CreatePartitionsStepDefinition().run(partition_args).partition_ids

    def build_dataset(self, partition_ids: list[str]) -> list[DatasetBuildTaskResults]:
        """Further partition the dataset into windows and build the rslearn dataset for the given partition"""
        logger.info(f"Building dataset for {len(partition_ids)} partitions")
        results: list[DatasetBuildTaskResults] = []
        for partition_id in partition_ids:
            dataset_build_args = DatasetBuildTaskArgs(
                partition_ids=[partition_id],
                scratch_path=str(self.scratch_path),
                model_stage_id=self.model_stage_id,
                model_stage_root_path=self.model_stage_paths.root_path,
                dataset_path=self.dataset_path,
            )
            results.append(DatasetBuildStepDefinition().run(dataset_build_args))
        return results

    def run_inference(self, partition_id: str) -> RunInferenceTaskResults:
        """Run inference on the given partition"""
        logger.debug(f"Running inference step for partition {partition_id}")
        inference_args = RunInferenceTaskArgs(
            partition_ids=[partition_id],
            scratch_path=str(self.scratch_path),
            model_stage_id=self.model_stage_id,
            model_stage_root_path=self.model_stage_paths.root_path,
            dataset_path=self.dataset_path,
        )
        print(f'inference_args: {inference_args}')
        print(f'RunInferenceStepDefinition: {RunInferenceStepDefinition()}')
        results = RunInferenceStepDefinition().run(inference_args)
        print(f'results: {results}')
        self.inference_results_data_type = results.inference_results_data_type

        return results

    def postprocess(self, partition_id: str) -> PostprocessPartitionTaskResults:
        """Postprocess the results of inference for the given partition"""
        if self.inference_results_data_type is None:
            raise ValueError("inference_results_data_type must be set before postprocessing")

        postprocess_args = PostprocessPartitionTaskArgs(
            partition_ids=[partition_id],
            scratch_path=str(self.scratch_path),
            model_stage_id=self.model_stage_id,
            model_stage_root_path=self.model_stage_paths.root_path,
            dataset_path=self.dataset_path,
        )
        return PostprocessPartitionStepDefinition().run(postprocess_args)

    def combine(self, partition_ids: list[str]) -> CombinePartitionsTaskResults:
        """Combine the results of inference from all partitions into a single result set"""
        if self.inference_results_data_type is None:
            raise ValueError("inference_results_data_type must be set before combining")

        combine_args = CombinePartitionsTaskArgs(
            partition_ids=partition_ids,
            scratch_path=str(self.scratch_path),
            model_stage_id=self.model_stage_id,
            model_stage_root_path=self.model_stage_paths.root_path,
            dataset_path=self.dataset_path,
        )
        return CombinePartitionsStepDefinition().run(combine_args)

    def run_pipeline(self, steps: list[str], partitions: list[str] | None = None) -> None:
        """Run the pipeline using the provided steps and partitions."""
        if 'partition' in steps:
            logger.info('Running partition step')
            if partitions:
                logger.warning('Partitions were provided, but the partition step will re-create them anyway.')
            partitions = self.partition()
            logger.info(f'Created {len(partitions)} partitions')
        else:
            logger.info('Fetching existing partitions')
            partitions = partitions or self.scratch.get_partitions()
            logger.info(f'Found {len(partitions)} partitions')

        if not partitions:
            logger.error("No partitions found to process. You may have forgotten to: "
                         "  - Previously run the `partition` step, "
                         "  - Pass `partition` as a step"
                         "  - Pass the partition-ids argument.")
            sys.exit(1)

        for step in steps:
            if step == 'dataset_build':
                logger.info(f'Running dataset build step for {len(partitions)} partitions')
                all(self.build_dataset(partitions))
                logger.info('Completed dataset build step')
            elif step == 'inference':
                for partition_id in partitions:
                    self.run_inference(partition_id)
            elif step == 'postprocess':
                for partition_id in partitions:
                    self.postprocess(partition_id)

        if 'combine' in steps:
            logger.info(f'Running combine step for {len(partitions)} partitions')
            self.combine(partitions)


def main() -> None:
    import argparse
    from olmoearth_shared.tools.telemetry.logging_tools import configure_logging

    parser = argparse.ArgumentParser(description="Earth System Run Local Prediction Runner")
    parser.add_argument("project_path", nargs='?', default=str(Path.cwd()), type=Path,
                       help="Path to project directory")
    parser.add_argument("--scratch-path",
                       help="Scratch directory path (default: {project_path}/scratch)")
    parser.add_argument("--request-geometry-path", type=Path,
                       help="Path to the input prediction_request_geometry.geojson file",
                        default=Path(PREDICTION_REQUEST_GEOMETRY_FILE_NAME))
    parser.add_argument("--steps",
                       help="Comma-separated steps to run (default: all steps)")
    parser.add_argument("--partition-ids",
                       help="Comma-separated partition IDs to run steps on")

    args = parser.parse_args()

    configure_logging()

    request_geometry_path = Path(args.request_geometry_path).absolute()
    project_path = Path(args.project_path).absolute()
    scratch_path = Path(args.scratch_path).absolute() if args.scratch_path else project_path / "scratches" / request_geometry_path.stem
    scratch_path.mkdir(parents=True, exist_ok=True)

    runner = OlmoEarthRunPredictRunner(
        project_path=project_path,
        scratch_path=scratch_path,
        request_geometry_path=request_geometry_path,
    )

    allowed_steps = ["partition", "dataset_build", "inference", "postprocess", "combine"]
    steps = args.steps.split(",") if args.steps else allowed_steps
    if len(set(steps).difference(allowed_steps)) > 0:
        logger.error(f"Invalid steps provided. Allowed steps are: {', '.join(allowed_steps)}")
        sys.exit(1)

    partition_ids = args.partition_ids.split(",") if args.partition_ids else None
    logger.info(f"Running steps ({','.join(steps)}) on partitions: {partition_ids}")
    runner.run_pipeline(steps, partition_ids)


if __name__ == "__main__":
    main()



import importlib
from typing import Any

from upath import UPath

from olmoearth_run.runner.tools.partitioners.partition_interface import PartitionInterface
from olmoearth_run.runner.tools.postprocessors.postprocess_interface import PostprocessInterfaceRaster, PostprocessInterfaceVector
from olmoearth_run.runner.tools.labeled_window_preparers.labeled_window_preparer import LabeledWindowPreparer
from olmoearth_run.runner.tools.data_splitters.data_splitter_interface import DataSplitterInterface
from olmoearth_run.runner.tools.samplers.sampler_interface import SamplerInterface
from olmoearth_run.runner.tools.samplers.noop_sampler import NoopSampler
from olmoearth_run.shared.models.olmoearth_run_config import OlmoEarthRunConfig


class OlmoEarthRunConfigLoader:
    @staticmethod
    def load_olmoearth_run_config(yaml_path: str) -> OlmoEarthRunConfig:
        """Load OlmoEarthRun configuration from a YAML file."""
        with UPath(yaml_path).open('r') as file:
            yaml_content = file.read()
        return OlmoEarthRunConfig.from_yaml(yaml_content)

    # Window preparation methods (training phase)
    @classmethod
    def get_labeled_window_preparer(cls, config: OlmoEarthRunConfig) -> LabeledWindowPreparer:
        """Loads and initializes a labeled window preparer based on the OlmoEarthRun configuration."""
        if config.window_prep is None:
            raise ValueError("window_prep configuration is required but not found in OlmoEarthRun config")
        preparer_config = config.window_prep.labeled_window_preparer
        preparer = cls._instantiate_from_dict(preparer_config)
        cls._validate_instance_type(preparer, LabeledWindowPreparer, "LabeledWindowPreparer")
        return preparer

    @classmethod
    def get_data_splitter(cls, config: OlmoEarthRunConfig) -> DataSplitterInterface:
        """Loads and initializes a data splitter based on the OlmoEarthRun configuration."""
        if config.window_prep is None:
            raise ValueError("window_prep configuration is required but not found in OlmoEarthRun config")
        splitter_config = config.window_prep.data_splitter
        splitter = cls._instantiate_from_dict(splitter_config)
        cls._validate_instance_type(splitter, DataSplitterInterface, "DataSplitterInterface")
        return splitter

    @classmethod
    def get_sampler(cls, config: OlmoEarthRunConfig) -> SamplerInterface:
        """Loads and initializes a sampler based on the OlmoEarthRun configuration."""
        if config.window_prep is None:
            raise ValueError("window_prep configuration is required but not found in OlmoEarthRun config")

        # If no sampler config provided, return NoopSampler
        if config.window_prep.sampler is None:
            return NoopSampler()

        sampler_config = config.window_prep.sampler
        sampler = cls._instantiate_from_dict(sampler_config)
        cls._validate_instance_type(sampler, SamplerInterface, "SamplerInterface")
        return sampler

    # Partition strategy methods (inference phase)
    @classmethod
    def get_partition_request_geometry_partitioner(cls, config: OlmoEarthRunConfig) -> PartitionInterface:
        """Loads and initializes a partitioner for partitioning request geometry."""
        return cls._get_partitioner_from_config(config.partition_strategies.partition_request_geometry)

    @classmethod
    def get_prepare_window_geometries_partitioner(cls, config: OlmoEarthRunConfig) -> PartitionInterface:
        """Loads and initializes a partitioner for preparing window geometries."""
        return cls._get_partitioner_from_config(config.partition_strategies.prepare_window_geometries)

    # Postprocessor strategy methods (inference phase)
    @classmethod
    def get_window_postprocessor(cls, config: OlmoEarthRunConfig) -> PostprocessInterfaceRaster | PostprocessInterfaceVector:
        """Loads and initializes a postprocessor for window-level processing."""
        return cls._get_postprocessor_from_config(config.postprocessing_strategies.process_window)

    @classmethod
    def get_partition_postprocessor(cls, config: OlmoEarthRunConfig) -> PostprocessInterfaceRaster | PostprocessInterfaceVector:
        """Loads and initializes a postprocessor for partition-level processing."""
        return cls._get_postprocessor_from_config(config.postprocessing_strategies.process_partition)

    @classmethod
    def get_dataset_postprocessor(cls, config: OlmoEarthRunConfig) -> PostprocessInterfaceRaster | PostprocessInterfaceVector:
        """Loads and initializes a postprocessor for dataset-level processing."""
        return cls._get_postprocessor_from_config(config.postprocessing_strategies.process_dataset)

    @classmethod
    def _get_partitioner_from_config(cls, config_dict: dict[str, Any]) -> PartitionInterface:
        """Helper method to instantiate and validate a partitioner from config."""
        partitioner = cls._instantiate_from_dict(config_dict)
        cls._validate_instance_type(partitioner, PartitionInterface, "PartitionInterface")
        return partitioner

    @classmethod
    def _get_postprocessor_from_config(cls, config_dict: dict[str, Any]) -> PostprocessInterfaceRaster | PostprocessInterfaceVector:
        """Helper method to instantiate and validate a postprocessor from config."""
        postprocessor = cls._instantiate_from_dict(config_dict)
        cls._validate_instance_type(postprocessor, PostprocessInterfaceRaster | PostprocessInterfaceVector, "Postprocessor")
        return postprocessor

    @classmethod
    def _validate_instance_type(cls, obj: Any, expected_type: Any, type_name: str) -> None:
        """Helper method to validate that an instantiated object matches the expected type."""
        if not isinstance(obj, expected_type):
            raise ValueError(f"Initialized object was not a {type_name}. Got type: {type(obj)}")

    @classmethod
    def _instantiate_from_dict(cls, config: dict[str, Any]) -> Any:
        """
        Recursively instantiates an object from a dictionary configuration.
        This is the core logic for handling nested objects and _module_/_callable_ calls.
        """
        if "_module_" in config and "_callable_" in config:
            # Handle instantiation via a callable (static method, class method, function)
            callable = cls._get_callable_from_module_and_path(config["_module_"], config["_callable_"])

            # Collect arguments, excluding internal keys
            call_args = {k: cls._instantiate_from_dict(v) if isinstance(v, dict) else v for k, v in config.items() if k not in ["_module_", "_callable_"]}
            return callable(**call_args)

        if "class_path" in config:
            # Handle direct class instantiation
            module_name, class_name = config["class_path"].rsplit(".", 1)
            module = importlib.import_module(module_name)
            target_class = getattr(module, class_name)

            init_args = config.get("init_args", {})

            # Recursively instantiate nested init_args
            resolved_init_args = {k: cls._instantiate_from_dict(v) if isinstance(v, dict) else v for k, v in init_args.items()}

            return target_class(**resolved_init_args)

        # If no special keys, check for nested dictionaries that still need instantiation
        resolved_value = {}
        is_complex_dict = False
        for k, v in config.items():
            if isinstance(v, dict):
                resolved_value[k] = cls._instantiate_from_dict(v)
                is_complex_dict = True
            else:
                resolved_value[k] = v
        return resolved_value if is_complex_dict else config  # Return original if no nested dicts were found

    @classmethod
    def _get_callable_from_module_and_path(cls, module_name: str, callable_path: str) -> Any:
        """Imports the specified module and then resolves the callable path within that module"""
        module = importlib.import_module(module_name)

        current_obj = module
        parts = callable_path.split(".")
        for part in parts:
            current_obj = getattr(current_obj, part)
        return current_obj



"""
Local runner for the OlmoEarth prediction pipeline using unified config.

This runner uses the unified OlmoEarthConfig format (olmoearth_config.yaml) for configuration.
The step definitions automatically detect unified config and use PredictionToolsClassLoader
to instantiate partitioners and postprocessors.
"""

import logging
import shutil
import sys
import uuid
from os import PathLike
from pathlib import Path

from olmoearth_run.runner.steps.combine_partitions_step_definition import (
    CombinePartitionsStepDefinition,
)
from olmoearth_run.runner.steps.create_partitions_step_definition import (
    CreatePartitionsStepDefinition,
)
from olmoearth_run.runner.steps.dataset_build_step_definition import (
    DatasetBuildStepDefinition,
)
from olmoearth_run.runner.steps.postprocess_partition_step_definition import (
    PostprocessPartitionStepDefinition,
)
from olmoearth_run.runner.steps.run_inference_step_definition import (
    RunInferenceStepDefinition,
)
from olmoearth_run.shared.models.api.task_args import (
    CombinePartitionsTaskArgs,
    CreatePartitionsTaskArgs,
    DatasetBuildTaskArgs,
    PostprocessPartitionTaskArgs,
    RunInferenceTaskArgs,
)
from olmoearth_run.shared.models.api.task_results import (
    CombinePartitionsTaskResults,
    DatasetBuildTaskResults,
    InferenceResultsDataType,
    PostprocessPartitionTaskResults,
    RunInferenceTaskResults,
)
from olmoearth_run.shared.models.model_stage_paths import (
    CHECKPOINT_FILE_NAME,
    ModelStagePaths,
)
from olmoearth_run.shared.models.prediction_scratch_space import (
    OLMOEARTH_CONFIG_FILE_NAME,
    PREDICTION_REQUEST_GEOMETRY_FILE_NAME,
    PredictionScratchSpace,
)

logger = logging.getLogger(__name__)

NAMESPACE_UNIFIED_PREDICT_RUNNER = uuid.UUID("0f94b556-2bc6-40c2-a8ae-e36d4ab0cf03")


class UnifiedConfigPredictRunner:
    """
    Local runner for the OlmoEarth prediction pipeline.

    Uses the unified OlmoEarthConfig format (olmoearth_config.yaml) for configuration.
    """

    def __init__(
        self,
        project_path: PathLike,
        scratch_path: PathLike,
        request_geometry_path: PathLike | None = None,
        checkpoint_path: PathLike | None = None,
    ):
        self.project_path = Path(project_path)
        self.scratch_path = Path(scratch_path)
        self.dataset_path = f"{self.scratch_path}/dataset_0"

        self.stage_name = "model_stage_0"
        self.stage_path = self.scratch_path / self.stage_name
        self.project_name = self.project_path.name
        self.inference_results_data_type: InferenceResultsDataType | None = None

        self.scratch = PredictionScratchSpace(root_path=str(self.scratch_path))
        self.model_stage_paths = ModelStagePaths(root_path=str(self.stage_path))

        self.project_uuid_namespace = uuid.uuid5(
            NAMESPACE_UNIFIED_PREDICT_RUNNER, self.project_name
        )
        self.model_stage_id = uuid.uuid5(self.project_uuid_namespace, self.stage_name)

        self.checkpoint_path = Path(
            checkpoint_path or self.project_path / CHECKPOINT_FILE_NAME
        )
        self.request_geometry_path = Path(
            request_geometry_path
            or self.project_path / PREDICTION_REQUEST_GEOMETRY_FILE_NAME
        )

        logger.info(f"Project Path: {self.project_path}")
        logger.info(f"Scratch Path: {self.scratch_path}")
        logger.info(f"Dataset Path: {self.dataset_path}")

        self._setup_project_env()

    def _setup_project_env(self) -> None:
        """Set up the project environment in the scratch space by copying config and creating symlinks."""
        Path(self.scratch_path).mkdir(parents=True, exist_ok=True)
        Path(self.dataset_path).mkdir(exist_ok=True)

        # Copy olmoearth_config.yaml to scratch space
        # Step definitions detect this file and use unified config automatically
        olmoearth_config_path = self.project_path / OLMOEARTH_CONFIG_FILE_NAME
        if not olmoearth_config_path.exists():
            logger.error(
                f"Required project file not found: {olmoearth_config_path}. "
                "Expected unified OlmoEarthConfig format (olmoearth_config.yaml)."
            )
            sys.exit(1)

        scratch_config_path = Path(self.scratch.olmoearth_config_path)
        if not scratch_config_path.exists():
            logger.info(f"Copying {olmoearth_config_path} to {scratch_config_path}")
            scratch_config_path.parent.mkdir(parents=True, exist_ok=True)
            shutil.copy2(olmoearth_config_path, scratch_config_path)
        else:
            logger.info(f"{scratch_config_path} already exists. Skipping copy.")

        # Set up symlinks for checkpoint and request geometry
        symlink_files: list[tuple[Path, Path, bool]] = [
            # (target, link, is_optional)
            (
                self.request_geometry_path,
                Path(self.scratch.prediction_request_geometry_path),
                False,
            ),
            (self.checkpoint_path, Path(self.model_stage_paths.checkpoint_path), True),
        ]

        for target, link, is_optional in symlink_files:
            if not link.exists():
                if not target.exists():
                    if is_optional:
                        logger.warning(
                            f"Optional file {target} does not exist. Skipping linking."
                        )
                        continue
                    else:
                        logger.error(f"Required project file {target} does not exist.")
                        sys.exit(1)
                logger.info(f"{link} does not exist yet. Linking {target} to {link}")
                link.parent.mkdir(parents=True, exist_ok=True)
                link.symlink_to(target)
            else:
                logger.info(
                    f"{link} already exists. Skipping linking {target} to {link}"
                )

    def partition(self) -> list[str]:
        """Partition the original request feature into smaller features for processing"""
        partition_args = CreatePartitionsTaskArgs(
            scratch_path=str(self.scratch_path),
            model_stage_root_path=self.model_stage_paths.root_path,
            model_stage_id=self.model_stage_id,
            dataset_path=self.dataset_path,
        )
        return CreatePartitionsStepDefinition().run(partition_args).partition_ids

    def build_dataset(self, partition_ids: list[str]) -> list[DatasetBuildTaskResults]:
        """Further partition the dataset into windows and build the rslearn dataset for the given partition"""
        logger.info(f"Building dataset for {len(partition_ids)} partitions")
        results: list[DatasetBuildTaskResults] = []
        for partition_id in partition_ids:
            dataset_build_args = DatasetBuildTaskArgs(
                partition_ids=[partition_id],
                scratch_path=str(self.scratch_path),
                model_stage_id=self.model_stage_id,
                model_stage_root_path=self.model_stage_paths.root_path,
                dataset_path=self.dataset_path,
            )
            results.append(DatasetBuildStepDefinition().run(dataset_build_args))
        return results

    def run_inference(self, partition_id: str) -> RunInferenceTaskResults:
        """Run inference on the given partition"""
        logger.debug(f"Running inference step for partition {partition_id}")
        inference_args = RunInferenceTaskArgs(
            partition_ids=[partition_id],
            scratch_path=str(self.scratch_path),
            model_stage_id=self.model_stage_id,
            model_stage_root_path=self.model_stage_paths.root_path,
            dataset_path=self.dataset_path,
        )
        results = RunInferenceStepDefinition().run(inference_args)
        self.inference_results_data_type = results.inference_results_data_type

        return results

    def postprocess(self, partition_id: str) -> PostprocessPartitionTaskResults:
        """Postprocess the results of inference for the given partition"""
        if self.inference_results_data_type is None:
            raise ValueError(
                "inference_results_data_type must be set before postprocessing"
            )

        postprocess_args = PostprocessPartitionTaskArgs(
            partition_ids=[partition_id],
            scratch_path=str(self.scratch_path),
            model_stage_id=self.model_stage_id,
            model_stage_root_path=self.model_stage_paths.root_path,
            dataset_path=self.dataset_path,
        )
        return PostprocessPartitionStepDefinition().run(postprocess_args)

    def combine(self, partition_ids: list[str]) -> CombinePartitionsTaskResults:
        """Combine the results of inference from all partitions into a single result set"""
        if self.inference_results_data_type is None:
            raise ValueError("inference_results_data_type must be set before combining")

        combine_args = CombinePartitionsTaskArgs(
            partition_ids=partition_ids,
            scratch_path=str(self.scratch_path),
            model_stage_id=self.model_stage_id,
            model_stage_root_path=self.model_stage_paths.root_path,
            dataset_path=self.dataset_path,
        )
        return CombinePartitionsStepDefinition().run(combine_args)

    def run_pipeline(
        self, steps: list[str], partitions: list[str] | None = None
    ) -> None:
        """Run the pipeline using the provided steps and partitions."""
        if "partition" in steps:
            logger.info("Running partition step")
            if partitions:
                logger.warning(
                    "Partitions were provided, but the partition step will re-create them anyway."
                )
            partitions = self.partition()
            logger.info(f"Created {len(partitions)} partitions")
        else:
            logger.info("Fetching existing partitions")
            partitions = partitions or self.scratch.get_partitions()
            logger.info(f"Found {len(partitions)} partitions")

        if not partitions:
            logger.error(
                "No partitions found to process. You may have forgotten to: "
                "  - Previously run the `partition` step, "
                "  - Pass `partition` as a step"
                "  - Pass the partition-ids argument."
            )
            sys.exit(1)

        for step in steps:
            if step == "dataset_build":
                logger.info(
                    f"Running dataset build step for {len(partitions)} partitions"
                )
                all(self.build_dataset(partitions))
                logger.info("Completed dataset build step")
            elif step == "inference":
                for partition_id in partitions:
                    self.run_inference(partition_id)
            elif step == "postprocess":
                for partition_id in partitions:
                    self.postprocess(partition_id)

        if "combine" in steps:
            logger.info(f"Running combine step for {len(partitions)} partitions")
            self.combine(partitions)


def main() -> None:
    import argparse

    from olmoearth_shared.tools.telemetry.logging_tools import configure_logging

    parser = argparse.ArgumentParser(
        description="OlmoEarth Unified Config Prediction Runner"
    )
    parser.add_argument(
        "project_path",
        nargs="?",
        default=str(Path.cwd()),
        type=Path,
        help="Path to project directory containing olmoearth_config.yaml",
    )
    parser.add_argument(
        "--scratch-path",
        help="Scratch directory path (default: {project_path}/scratches/{geometry_name})",
    )
    parser.add_argument(
        "--request-geometry-path",
        type=Path,
        help="Path to the input prediction_request_geometry.geojson file",
        default=Path(PREDICTION_REQUEST_GEOMETRY_FILE_NAME),
    )
    parser.add_argument(
        "--checkpoint-path", type=Path, help="Path to the model checkpoint file"
    )
    parser.add_argument(
        "--steps", help="Comma-separated steps to run (default: all steps)"
    )
    parser.add_argument(
        "--partition-ids", help="Comma-separated partition IDs to run steps on"
    )

    args = parser.parse_args()

    configure_logging()

    project_path = Path(args.project_path).absolute()
    # If request geometry path is just a filename, look for it in the project directory
    if args.request_geometry_path.parent == Path("."):
        request_geometry_path = (project_path / args.request_geometry_path).absolute()
    else:
        request_geometry_path = Path(args.request_geometry_path).absolute()
    scratch_path = (
        Path(args.scratch_path).absolute()
        if args.scratch_path
        else project_path / "scratches" / request_geometry_path.stem
    )
    scratch_path.mkdir(parents=True, exist_ok=True)
    checkpoint_path = (
        Path(args.checkpoint_path).absolute() if args.checkpoint_path else None
    )

    runner = UnifiedConfigPredictRunner(
        project_path=project_path,
        scratch_path=scratch_path,
        request_geometry_path=request_geometry_path,
        checkpoint_path=checkpoint_path,
    )

    allowed_steps = [
        "partition",
        "dataset_build",
        "inference",
        "postprocess",
        "combine",
    ]
    steps = args.steps.split(",") if args.steps else allowed_steps
    if len(set(steps).difference(allowed_steps)) > 0:
        logger.error(
            f"Invalid steps provided. Allowed steps are: {', '.join(allowed_steps)}"
        )
        sys.exit(1)

    partition_ids = args.partition_ids.split(",") if args.partition_ids else None
    logger.info(f"Running steps ({','.join(steps)}) on partitions: {partition_ids}")
    runner.run_pipeline(steps, partition_ids)


if __name__ == "__main__":
    main()



"""Utilities for loading and processing spatial masks for inference filtering."""
import json
import logging
from typing import Any, Literal, cast

import numpy as np
import rasterio  # type: ignore[import-untyped]
from affine import Affine  # type: ignore[import-untyped]
from rasterio.crs import CRS  # type: ignore[import-untyped]
from rasterio.session import GSSession, DummySession  # type: ignore[import-untyped]
from pydantic import BaseModel, Field
from rasterio.warp import reproject, Resampling, transform_bounds  # type: ignore[import-untyped]
from rasterio.features import geometry_mask  # type: ignore[import-untyped]
from rslearn.utils.geometry import Projection, STGeometry, WGS84_PROJECTION, WGS84_EPSG
from shapely import box
from shapely.geometry import shape, mapping
from shapely.geometry.base import BaseGeometry
from shapely.ops import unary_union, transform as shapely_transform
from shapely.prepared import prep
from pyproj import Transformer

from olmoearth_shared.api.run.prediction_geometry import PredictionResultCollection
from olmoearth_shared.tools.gcs_tools import (
    is_gcs_path,
    list_gcs_directory,
    read_text_from_gcs,
)

logger = logging.getLogger(__name__)


class MaskConfig:
    """Configuration for spatial masking, bundling mask files with their valid values."""

    def __init__(self, mask_files: list["MaskFileInfo"], valid_values: list[int]):
        self.mask_files = mask_files
        self.valid_values = valid_values

    @classmethod
    def from_gcs_path(cls, gcs_path: str | None, valid_values: list[int]) -> "MaskConfig | None":
        """Load mask config from a GCS path. Returns None if gcs_path is None."""
        if not gcs_path:
            return None
        mask_files = MaskFileInfo.load_mask_files(gcs_path)
        return cls(mask_files, valid_values)


class MaskFileInfo(BaseModel):
    """Information about a mask file."""

    model_config = {"arbitrary_types_allowed": True}

    path: str = Field(description="Path to the mask file (GCS or local)")
    bounding_box: BaseGeometry = Field(description="Bounding box in WGS84")
    file_type: Literal["geotiff", "geojson"] = Field(description="Type of mask file")

    def to_stgeometry(self) -> STGeometry:
        """Returns the mask bounding box as an STGeometry in WGS84."""
        return STGeometry(WGS84_PROJECTION, self.bounding_box, time_range=None)

    @staticmethod
    def load_mask_files(gcs_path: str) -> list["MaskFileInfo"]:
        """Load mask file infos from a GCS path (single file or directory)."""
        mask_files = MaskFileInfo._list_mask_files(gcs_path)
        logger.info(f"Loading {len(mask_files)} mask files from {gcs_path}")
        return [MaskFileInfo._get_mask_file_info(mask_file) for mask_file in mask_files]

    @staticmethod
    def _list_mask_files(gcs_path: str) -> list[str]:
        """List mask files from a GCS path (single file or directory)."""
        if gcs_path.endswith(".tif") or gcs_path.endswith(".geojson"):
            return [gcs_path]

        # It's a directory - list files
        all_files = list_gcs_directory(gcs_path)
        mask_files = [f for f in all_files if f.endswith(".tif") or f.endswith(".geojson")]

        if not mask_files:
            raise ValueError(f"No mask files (.tif or .geojson) found in {gcs_path}")

        # Validate all files are same type
        tif_files = [f for f in mask_files if f.endswith(".tif")]
        geojson_files = [f for f in mask_files if f.endswith(".geojson")]

        if tif_files and geojson_files:
            raise ValueError(f"Mixed mask types. Found {len(tif_files)} .tif and {len(geojson_files)} .geojson files.")

        return mask_files

    @staticmethod
    def _get_geotiff_info(file_path: str) -> "MaskFileInfo":
        """Get mask info from a GeoTIFF file by reading only the header. Reads directly from GCS."""
        session = GSSession() if is_gcs_path(file_path) else DummySession()
        with rasterio.Env(session=session), rasterio.open(file_path) as src:
            # Convert bounds to WGS84
            wgs84_bounds = transform_bounds(src.crs, CRS.from_epsg(WGS84_EPSG), *src.bounds)
            wgs84_box = box(*wgs84_bounds)
            return MaskFileInfo(
                path=file_path,
                bounding_box=wgs84_box,
                file_type="geotiff",
            )

    @staticmethod
    def _get_geojson_info(mask_path: str) -> "MaskFileInfo":
        """Get mask info from GeoJSON text content."""
        geometries = get_shapes_from_geojson_file(mask_path)
        if not geometries:
            raise ValueError(f"No geometries found in GeoJSON file: {mask_path}")

        combined = unary_union(geometries)

        # GeoJSON is already WGS84 - convert to bounding box
        return MaskFileInfo(path=mask_path, bounding_box=box(*combined.bounds), file_type="geojson")

    @staticmethod
    def _get_mask_file_info(mask_path: str) -> "MaskFileInfo":
        """Get mask info for a single file."""
        if mask_path.endswith(".tif"):
            return MaskFileInfo._get_geotiff_info(mask_path)
        elif mask_path.endswith(".geojson"):
            return MaskFileInfo._get_geojson_info(mask_path)
        else:
            raise ValueError(f"Unsupported mask file type: {mask_path}")


def get_shapes_from_geojson_file(geojson_path: str) -> list[BaseGeometry]:
    if is_gcs_path(geojson_path):
        geojson_text = read_text_from_gcs(geojson_path)
    else:
        with open(geojson_path, "r") as f:
            geojson_text = f.read()
    geojson = json.loads(geojson_text)

    if geojson.get("type") == "FeatureCollection":
        geometries = [shape(f["geometry"]) for f in geojson["features"] if f.get("geometry")]
    elif geojson.get("type") == "Feature":
        geometries = [shape(geojson["geometry"])]
    else:
        geometries = [shape(geojson)]
    return geometries


def get_first_intersecting_mask_file(mask_infos: list[MaskFileInfo], geometry: STGeometry) -> MaskFileInfo | None:
    """Get the first mask file whose bounding box intersects with the given geometry."""
    geometry_wgs84 = geometry.to_projection(WGS84_PROJECTION)
    for info in mask_infos:
        if info.bounding_box.intersects(geometry_wgs84.shp):
            return info
    return None


def get_first_intersecting_mask_for_raster(
    mask_infos: list[MaskFileInfo],
    bounds: tuple[float, float, float, float],
    crs: CRS,
) -> MaskFileInfo | None:
    """Get the first mask file whose bounds intersect with the given raster bounds.

    Args:
        mask_infos: List of mask file info objects to search
        bounds: Raster bounds as (left, bottom, right, top)
        crs: CRS of the raster bounds
    """
    if not mask_infos:
        return None
    wgs84_bounds = transform_bounds(crs, CRS.from_epsg(WGS84_EPSG), *bounds)
    raster_geom = box(*wgs84_bounds)
    for info in mask_infos:
        if info.bounding_box.intersects(raster_geom):
            return info
    return None


def _window_intersects_raster_mask(mask_file: MaskFileInfo, window_geometry: STGeometry, valid_values: list[int]) -> bool:
    """Check if a GeoTIFF mask has valid pixels within window bounds. Reads directly from GCS."""
    session = GSSession() if is_gcs_path(mask_file.path) else DummySession()
    with rasterio.Env(session=session), rasterio.open(mask_file.path) as src:
        # Transform query geometry to mask CRS
        mask_projection = Projection(src.crs, x_resolution=1, y_resolution=1)
        query_in_mask_crs = window_geometry.to_projection(mask_projection)
        query_bounds = cast(BaseGeometry, query_in_mask_crs.shp).bounds

        # Create a window from the bounds
        window = rasterio.windows.from_bounds(*query_bounds, src.transform)

        # Clamp window to valid range
        window = window.intersection(rasterio.windows.Window(0, 0, src.width, src.height))

        if window.width <= 0 or window.height <= 0:
            return False

        # Read the data in the window
        data = src.read(1, window=window)

        # Check if any valid values exist
        return bool(np.isin(data, valid_values).any())


def _window_intersects_vector_mask(mask_file: MaskFileInfo, window_geometry: STGeometry) -> bool:
    """Check if a GeoJSON mask intersects with geometry. Reads directly from GCS."""
    geometries = get_shapes_from_geojson_file(mask_file.path)

    # GeoJSON is always WGS84 - convert query geometry to WGS84
    query_wgs84 = window_geometry.to_projection(WGS84_PROJECTION)
    return any(geom.intersects(query_wgs84.shp) for geom in geometries)


def window_intersects_valid_mask(window_geometry: STGeometry, mask_info: MaskFileInfo | None, valid_values: list[int]) -> bool:
    """
    Check if a window intersects with valid pixels in the mask.
    This method is used to determine if we should inference on a window at all.
    """
    if mask_info is None:
        return False

    if mask_info.file_type == "geotiff":
        return _window_intersects_raster_mask(mask_info, window_geometry, valid_values)
    else:
        return _window_intersects_vector_mask(mask_info, window_geometry)


def apply_mask_to_raster_data(
    data: np.ndarray,
    transform: Affine,
    crs: CRS,
    mask_info: MaskFileInfo | None,
    valid_values: list[int],
    nodata_value: int | float = 0,
) -> None:
    """Apply mask to numpy array in-place, setting pixels outside valid mask areas to nodata."""
    # Determine array shape (handles both 2D and 3D arrays)
    if data.ndim == 3:
        height, width = data.shape[1], data.shape[2]
    else:
        height, width = data.shape[0], data.shape[1]

    if mask_info is None:
        # No mask - set entire array to nodata
        data[:] = nodata_value
        return

    # Build mask array (True = valid, False = masked)
    if mask_info.file_type == "geotiff":
        mask_array = _reproject_geotiff_mask(mask_info, transform, crs, height, width, valid_values)
    else:
        mask_array = _rasterize_geojson_mask(mask_info, transform, crs, height, width)

    # Apply mask to array
    # `mask_array` is (height, width). For 3D rasters we use bands-first (bands, height, width),
    # so `data[:, ~mask_array]` selects all bands at the invalid pixels.
    if data.ndim == 3:
        # Bands-first: (bands, height, width)
        data[:, ~mask_array] = nodata_value
    else:
        data[~mask_array] = nodata_value


def _reproject_geotiff_mask(
    mask_info: MaskFileInfo,
    dst_transform: Affine,
    dst_crs: CRS,
    height: int,
    width: int,
    valid_values: list[int],
) -> np.ndarray:
    """Load  a GeoTIFF mask and return the re-projected array of valid pixels"""
    mask_path = mask_info.path
    session = GSSession() if is_gcs_path(mask_path) else DummySession()
    with rasterio.Env(session=session), rasterio.open(mask_path) as src:
        mask_data = np.zeros((height, width), dtype=src.dtypes[0])

        reproject(
            source=rasterio.band(src, 1),
            destination=mask_data,
            src_transform=src.transform,
            src_crs=src.crs,
            dst_transform=dst_transform,
            dst_crs=dst_crs,
            resampling=Resampling.nearest,
        )

        result_mask = np.zeros((height, width), dtype=bool)
        for val in valid_values:
            result_mask |= (mask_data == val)
        return result_mask


def _rasterize_geojson_mask(
    mask_info: MaskFileInfo,
    dst_transform: Affine,
    dst_crs: CRS,
    height: int,
    width: int,
) -> np.ndarray:
    """Apply a GeoJSON mask and return the mask array."""
    geometries = get_shapes_from_geojson_file(mask_info.path)

    # Transform geometries from WGS84 to raster CRS
    transformer = Transformer.from_crs(f"EPSG:{WGS84_EPSG}", dst_crs, always_xy=True)
    transformed_geometries = [
        mapping(shapely_transform(transformer.transform, shp))
        for shp in geometries
    ]

    return geometry_mask(transformed_geometries, out_shape=(height, width), transform=dst_transform, invert=True)


def filter_features_by_mask(
    feature_collection: PredictionResultCollection,
    mask_info: MaskFileInfo | None,
    valid_values: list[int],
) -> PredictionResultCollection:
    """Filter a FeatureCollection to only include features that intersect with the mask."""
    if mask_info is None:
        return feature_collection

    # Convert features to shapely geometries once
    features_with_geoms: list[tuple[Any, BaseGeometry]] = []
    for feature in feature_collection.features:
        if feature.geometry is not None:
            features_with_geoms.append((feature, shape(feature.geometry.model_dump())))

    matched_features: list[Any] = []

    if mask_info.file_type == "geojson":
        # Vector mask - load geometries once and prepare for efficient intersection
        mask_geometries = get_shapes_from_geojson_file(mask_info.path)
        combined_mask = prep(unary_union(mask_geometries))
        for feature, geom in features_with_geoms:
            if combined_mask.intersects(geom):
                matched_features.append(feature)
    else:
        # Raster mask - open file once, check all features
        session = GSSession() if is_gcs_path(mask_info.path) else DummySession()
        with rasterio.Env(session=session), rasterio.open(mask_info.path) as src:
            transformer = Transformer.from_crs(f"EPSG:{WGS84_EPSG}", src.crs, always_xy=True)

            for feature, geom in features_with_geoms:
                geom_in_mask_crs = shapely_transform(transformer.transform, geom)
                query_bounds = geom_in_mask_crs.bounds

                window = rasterio.windows.from_bounds(*query_bounds, src.transform)
                window = window.intersection(rasterio.windows.Window(0, 0, src.width, src.height))

                if window.width <= 0 or window.height <= 0:
                    continue

                data = src.read(1, window=window)
                if np.isin(data, valid_values).any():
                    matched_features.append(feature)

    return PredictionResultCollection(type="FeatureCollection", features=matched_features)



"""Workflow notification models for callbacks."""
from typing import Annotated

from pydantic import BaseModel, Field

from olmoearth_shared.api.run.prediction_results import PredictionResultResponse
from olmoearth_shared.api.run.workflow import (
    DatasetBuildFromWindowsWorkflowResult,
    FineTuningWorkflowResponse,
    WorkflowResponse,
)

WorkflowResultInfo = Annotated[
    FineTuningWorkflowResponse | PredictionResultResponse | DatasetBuildFromWindowsWorkflowResult,
    Field(discriminator='workflow_type')
]


class WorkflowNotification(BaseModel):
    """Structure of the payload sent to clients when emitting Workflow status update notifications"""

    workflow: WorkflowResponse = Field(description="The Workflow that triggered the notification")
    result_info: WorkflowResultInfo | None = Field(
        default=None,
        description="If the workflow is complete, this contains the workflow-specific results"
    )
    error_message: str | None = Field(
        default=None,
        description="Error details if the Workflow failed",
    )



from enum import StrEnum
from typing import Annotated, Literal

from pydantic import BaseModel, Field
from typing_extensions import TypeAliasType


class PostprocessorName(StrEnum):
    NOOP = "noop"
    COMBINE_GEOTIFF = "combine_geotiff"


class NoopPostprocessor(BaseModel):
    """Postprocessor that does nothing."""
    name: Literal[PostprocessorName.NOOP]


class LinearValueTransformConfig(BaseModel):
    """Linear transform: output = scale * input + offset.

    Uses discriminated union pattern to support other transform types in the future.
    """
    type: Literal["linear"] = "linear"
    scale: float = 1.0
    offset: float = 0.0
    output_dtype: Literal["uint8", "int16", "int32", "float32"] | None = Field(
        default=None,
        description="The output dtype to transform to. If not specified, the output dtype will be the same as the input dtype."
    )


# Discriminated union for future extensibility (e.g., LogTransformConfig, etc.)
ValueTransformConfig = Annotated[LinearValueTransformConfig, Field(discriminator="type")]


class CombineGeotiff(BaseModel):
    """Postprocessor that combines the geotiffs into a single geotiff."""
    name: Literal[PostprocessorName.COMBINE_GEOTIFF]
    max_pixels_per_dimension: int = 10_000
    value_transform: ValueTransformConfig | None = None


Postprocessor = TypeAliasType(
    "Postprocessor",
    Annotated[
        NoopPostprocessor | CombineGeotiff,
        Field(discriminator="name")
    ]
)


class Postprocessors(BaseModel):
    window: Postprocessor = Field(description="Controls the postprocessing of the prediction results for a single window.")
    partition: Postprocessor = Field(description="Controls the postprocessing of all the window results for a single partition.")
    request: Postprocessor = Field(description="Controls the postprocessing of all the partition results for the entire request.")



from typing import cast

from shapely.geometry import shape, MultiPolygon as ShapelyMultiPolygon, Polygon as ShapelyPolygon
from shapely.ops import unary_union

# Pydantic v2
from geojson_pydantic import FeatureCollection
from geojson_pydantic.geometries import Polygon, MultiPolygon

def fc_to_multipolygon(fc: FeatureCollection) -> ShapelyMultiPolygon:
    """
    Convert a geojson-pydantic FeatureCollection of Polygon/MultiPolygon
    features into a single Shapely MultiPolygon.
    """
    polys: list[ShapelyPolygon] = []

    for f in fc.features:
        if f.geometry is None:
            continue

        # Accept Polygon or MultiPolygon (skip anything else)
        if not isinstance(f.geometry, (Polygon, MultiPolygon)):
            continue

        # shapely.shape expects a GeoJSON-like dict
        geom = shape(f.geometry.model_dump())

        # Normalize to polygons
        if geom.geom_type == "Polygon":
            polys.append(geom)  # type: ignore[arg-type]
        elif geom.geom_type == "MultiPolygon":
            assert isinstance(geom, ShapelyMultiPolygon)
            polys.extend(list(geom.geoms))  # type: ignore[assignment]

    if not polys:
        return ShapelyMultiPolygon([])

    # We want to dissolve touching/overlapping polygons into pieces so we use unary_union and
    # normalize the result back to MultiPolygon.
    dissolved = unary_union(polys)
    if dissolved.geom_type == "Polygon":
        assert isinstance(dissolved, ShapelyPolygon)
        return ShapelyMultiPolygon([dissolved])
    elif dissolved.geom_type == "MultiPolygon":
        assert isinstance(dissolved, ShapelyMultiPolygon)
        return cast(ShapelyMultiPolygon, dissolved)
    else:
        # (e.g., GeometryCollection containing polygons)
        return ShapelyMultiPolygon([g for g in getattr(dissolved, "geoms", []) if g.geom_type == "Polygon"])



from enum import StrEnum

from pydantic import BaseModel, Field

from olmoearth_run.config import OlmoEarthSettings
from olmoearth_run.shared.models.api.step_type import StepType
from olmoearth_run.shared.models.model_stage_paths import ModelStagePaths


class DataSource(StrEnum):
    PLANETARY_COMPUTER = "planetary_computer"


class DatasetOperationalContext(BaseModel):
    """Environment/context-specific information necessary for controlling behavior of dataset operations."""
    data_source: DataSource = Field(default=DataSource.PLANETARY_COMPUTER, description="The data source for acquiring imagery")
    dataset_path: str = Field(description="The path to the rslearn dataset to be used for training/validation/testing/inference")


class ModelOperationalContext(BaseModel):
    """Environment/context-specific information necessary for controlling behavior of dataset and model operations."""

    step_type: StepType = Field(description="The type of step being performed")
    dataset_path: str = Field(description="The path to the rslearn dataset to be used for training/validation/testing/inference")
    num_data_worker_processes: int = Field(description="The number of concurrent workers to use for data loading for lightning ops, or to perform dataset operations in parallel")
    batch_size: int = Field(default=1, description="The number of inputs to process in each forward pass.")
    wandb_name: str = Field(default="N/A", description="The name of the wandb run")
    model_stage_paths: ModelStagePaths = Field(description="The paths to the model stage directories")
    predict_groups: list[str] | None = Field(default=None, description="RSLearn groups to filter prediction to. Only used during inference. By default all groups are eligible for inference.")

    @property
    def is_training(self) -> bool:
        return self.step_type == StepType.FINE_TUNE

    @property
    def needs_targets(self) -> bool:
        """Returns True if this step requires target/label data to be loaded."""
        return self.step_type in (StepType.FINE_TUNE, StepType.MODEL_EVALUATION)

    @property
    def wandb_project_name(self) -> str:
        return OlmoEarthSettings.WANDB_PROJECT

    @property
    def wandb_entity(self) -> str:
        return OlmoEarthSettings.WANDB_ENTITY



import json
from collections.abc import Sequence
from typing import Any

from lightning.pytorch import LightningModule, Trainer
from lightning.pytorch.callbacks import BasePredictionWriter
from upath import UPath

from rslearn.train.model_context import ModelOutput


class EvaluationMetadataWriter(BasePredictionWriter):
    """Saves metadata during model evaluation.

    This callback captures the patch_bounds from each prediction batch
    and saves them to a JSON file. This allows us to later read predictions
    and labels at the exact locations used during inference for metric calculation.
    """

    def __init__(self, output_path: UPath):
        super().__init__(write_interval="batch")
        self.output_path = output_path
        self.metadata: dict[str, dict[str, Any]] = {}

    def write_on_batch_end(
        self,
        trainer: Trainer,
        pl_module: LightningModule,
        prediction: ModelOutput,
        batch_indices: Sequence[int] | None,
        batch: tuple[list, list, list],
        batch_idx: int,
        dataloader_idx: int,
    ) -> None:
        """Save metadata for each window in the batch."""
        _, _, metadatas = batch
        for metadata in metadatas:
            self.metadata[metadata.window_name] = {
                "patch_bounds": list(metadata.patch_bounds),
                "window_bounds": list(metadata.window_bounds),
                "window_group": metadata.window_group,
            }

    def on_predict_end(self, trainer: Trainer, pl_module: LightningModule) -> None:
        self.output_path.write_text(json.dumps(self.metadata, indent=2))



import uuid
from datetime import datetime
from enum import StrEnum
from typing import Literal, Self

from geojson_pydantic.geometries import Geometry
from pydantic import BaseModel, Field, model_validator

from olmoearth_shared.api.common.api_common_models import ApiOutputModel
from olmoearth_shared.api.common.search_filters import DatetimeFilter, KeywordFilter, NumericFilter, SortDirection, StringFilter
from olmoearth_shared.api.run.inference_results_config import InferenceResultsConfig
from olmoearth_shared.api.run.model_pipelines import ModelPipelineResponse
from olmoearth_shared.api.run.prediction_geometry import PredictionResultCollection, PredictionResultFeature
from olmoearth_shared.api.run.workflow import WorkflowResponse
from olmoearth_shared.api.run.workflow_type import WorkflowType


class PredictionResultFileProperties(BaseModel):
    """Properties for prediction output files."""

    filesize_bytes: int = Field(description="Size of the file in bytes")
    crs_code: str = Field(description="Coordinate Reference System code (e.g., EPSG:4326)")


class PredictionResultSortField(StrEnum):
    """Valid fields for sorting prediction results in search results."""
    START_DATETIME = "start_datetime"
    END_DATETIME = "end_datetime"
    CREATED_AT = "created_at"


class SearchPredictionResultsRequest(BaseModel):
    """Request model for searching prediction results"""
    id: KeywordFilter[uuid.UUID] | None = None
    model_pipeline_id: KeywordFilter[uuid.UUID] | None = None
    workflow_id: KeywordFilter[uuid.UUID] | None = None
    start_datetime: DatetimeFilter | None = None
    end_datetime: DatetimeFilter | None = None
    created_at: DatetimeFilter | None = None
    intersects_geometry: Geometry | None = Field(
        default=None,
        description="finds results whose geometry intersects with this geometry"
    )

    # Pagination and sorting
    limit: int = Field(default=50, ge=1, le=1000)
    offset: int = Field(default=0, ge=0)
    sort_by: PredictionResultSortField = Field(default=PredictionResultSortField.CREATED_AT)
    sort_direction: SortDirection = Field(default=SortDirection.DESC)


class PredictionResultFileSortField(StrEnum):
    """Valid fields for sorting prediction result files in search results."""
    FILE_NAME = "file_name"
    START_DATETIME = "start_datetime"
    END_DATETIME = "end_datetime"


class SearchPredictionResultFilesRequest(BaseModel):
    """Request model for searching prediction result files"""
    id: KeywordFilter[uuid.UUID] | None = None
    prediction_result_id: KeywordFilter[uuid.UUID] | None = None
    file_name: StringFilter | None = None
    mime_type: KeywordFilter[str] | None = None
    file_path: StringFilter | None = None
    start_datetime: DatetimeFilter | None = None
    end_datetime: DatetimeFilter | None = None
    intersects_geometry: Geometry | None = Field(
        default=None,
        description="Finds files whose geometry intersects with this geometry"
    )

    # Pagination and sorting
    limit: int = Field(default=50, ge=1, le=1000)
    offset: int = Field(default=0, ge=0)
    sort_by: PredictionResultFileSortField = Field(default=PredictionResultFileSortField.START_DATETIME)
    sort_direction: SortDirection = Field(default=SortDirection.DESC)


class PredictionResultFileResponse(BaseModel):
    """
    Response model for prediction result files.
    We always return this as part of the PredictionResult model below.
    """
    id: uuid.UUID
    prediction_result_id: uuid.UUID
    file_name: str
    mime_type: str
    geometry: Geometry
    file_path: str
    start_datetime: datetime
    end_datetime: datetime
    properties: PredictionResultFileProperties


class PredictionResultResponse(BaseModel):
    """
    Response model for prediction results.
    Include the Model Pipeline, Workflow, and all output files loaded.
    """
    workflow_type: Literal[WorkflowType.PREDICTION] = Field(default=WorkflowType.PREDICTION)
    id: uuid.UUID
    model_pipeline: ModelPipelineResponse
    workflow: WorkflowResponse
    geometry: Geometry
    start_datetime: datetime
    end_datetime: datetime
    inference_results_config: InferenceResultsConfig
    created_at: datetime
    output_files: list[PredictionResultFileResponse] = Field(default_factory=list)


class PropertyFilter(BaseModel):
    """
    When searching for GeoJSON features, this model allows the client to search on an arbitrary property
    in the GeoJSON feature's properties.
    """
    property_name: str = Field(description="name of field in geojson properties to search (e.g., 'name')")
    keyword_filter: KeywordFilter[str] | None = None
    string_filter: StringFilter | None = None
    numeric_filter: NumericFilter | None = None
    datetime_filter: DatetimeFilter | None = None

    @model_validator(mode='after')
    def has_exactly_one_filter(self) -> Self:
        filters = [self.keyword_filter, self.string_filter, self.numeric_filter, self.datetime_filter]
        if sum(f is not None for f in filters) != 1:
            raise ValueError("Exactly one filter must be provided")
        return self


class FeatureSortField(StrEnum):
    """Valid fields for sorting features in search results."""
    OE_CREATED_AT = "oe_created_at"
    OE_START_TIME = "oe_start_time"
    OE_END_TIME = "oe_end_time"


class SearchFeaturesRequest(BaseModel):
    """Request model for searching GeoJSON features stored in Elasticsearch."""

    id: KeywordFilter[str] | None = None
    intersects_geometry: Geometry | None = Field(
        default=None,
        description="Find features whose geometry intersects with this geometry"
    )

    # Known property filters (for convenience)
    oe_prediction_result_id: KeywordFilter[uuid.UUID] | None = None
    oe_prediction_result_file_id: KeywordFilter[uuid.UUID] | None = None
    oe_start_time: DatetimeFilter | None = None
    oe_end_time: DatetimeFilter | None = None
    oe_created_at: DatetimeFilter | None = None

    # Dynamic property filters for arbitrary properties
    property_filters: list[PropertyFilter] = Field(
        default_factory=list,
        description="Additional filters for arbitrary properties in the feature"
    )

    # Pagination and sorting
    limit: int = Field(default=50, ge=1, le=1000)
    offset: int = Field(default=0, ge=0)
    sort_by: FeatureSortField = FeatureSortField.OE_CREATED_AT
    sort_direction: SortDirection = SortDirection.DESC


# We can respond with either a standard ApiOutputModel or a Geojson Feature Collection
SearchFeaturesResponse = ApiOutputModel[PredictionResultFeature] | PredictionResultCollection


class PixelCoordinates(BaseModel):
    """Coordinate information for a pixel value query."""
    lat: float = Field(description="Latitude of the input point (WGS84)")
    lon: float = Field(description="Longitude of the input point (WGS84)")
    row: int = Field(description="Row index of the pixel in the raster")
    col: int = Field(description="Column index of the pixel in the raster")


class BandClassification(BaseModel):
    """Classification metadata for a band value."""
    label: str = Field(description="Human-readable label for the classification value")
    color: tuple[int, int, int, int] = Field(description="RGBA color values (0-255)")


class BandRegression(BaseModel):
    """Regression metadata for a band value."""
    min_value: float = Field(description="Minimum possible value for this field")
    max_value: float = Field(description="Maximum possible value for this field")
    colormap_name: str = Field(description="Name of the colormap used for visualization")


class BandValue(BaseModel):
    """Value and metadata for a single band at a pixel location."""
    band_index: int = Field(description="1-based band index in the raster")
    property_name: str = Field(description="Property name from inference results config")
    raw_value: float | int = Field(description="Raw pixel value from the raster")
    classification: BandClassification | None = Field(default=None, description="Classification metadata if applicable")
    regression: BandRegression | None = Field(default=None, description="Regression metadata if applicable")


class PixelValueRecord(BaseModel):
    """Complete pixel value information for a single prediction result file."""
    coordinates: PixelCoordinates = Field(description="Coordinate information")
    prediction_result_file: PredictionResultFileResponse = Field(description="File that provided this data")
    bands: list[BandValue] = Field(description="Band values and metadata")



from abc import ABC, abstractmethod
from typing import Generic

from olmoearth_run.runner.models.training.labeled_data import (
    AnnotationTask,
    TrainingWindowLabels,
    LabeledSTGeometry,
    LabeledWindow,
    RasterLabel,
)


class LabeledWindowPreparer(ABC, Generic[TrainingWindowLabels]):
    """Abstract base class for preparing labeled windows with generic label type."""

    @abstractmethod
    def prepare_labeled_windows(
        self,
        annotation_task: AnnotationTask,
    ) -> list[LabeledWindow[TrainingWindowLabels]]:
        """
        Prepare labeled windows from a single annotation task.

        The implementation should:
        1. Process the AnnotationTask to extract task context and annotations
        2. Transform annotation features into one or more training windows with labels
        3. Create one or more LabeledWindow objects with window properties and labels

        Args:
            annotation_task: Single AnnotationTask object containing task context and annotations

        Returns:
            List of LabeledWindow objects (can be 1 or more per task)
        """
        pass


class RasterLabelsWindowPreparer(LabeledWindowPreparer[list[RasterLabel]]):
    """Window preparer that produces raster (ndarray) labels."""

    @abstractmethod
    def prepare_labeled_windows(
        self,
        annotation_task: AnnotationTask,
    ) -> list[LabeledWindow[list[RasterLabel]]]:
        """Prepare labeled windows with raster labels."""
        pass


class VectorLabelsWindowPreparer(LabeledWindowPreparer[list[LabeledSTGeometry]]):
    """Window preparer that produces vector (LabeledSTGeometry list) labels."""

    @abstractmethod
    def prepare_labeled_windows(
        self,
        annotation_task: AnnotationTask,
    ) -> list[LabeledWindow[list[LabeledSTGeometry]]]:
        """Prepare labeled windows with vector labels."""
        pass



from pydantic import BaseModel, Field

from olmoearth_shared.models.olmoearth_config.labeled_data_prep.data_splitter import DataSplitter
from olmoearth_shared.models.olmoearth_config.labeled_data_prep.samplers import AnnotationSampler
from olmoearth_shared.models.olmoearth_config.labeled_data_prep.window_preparers import WindowPreparer


class LabeledDataPrep(BaseModel):
    """Controls the preparation of labeled data for training."""
    sampler: AnnotationSampler = Field(description="Controls the sampling of annotated data for training.")
    window_preparer: WindowPreparer = Field(description="Controls the preparation of annotated spatiotemporal geometries into labeled rslearn windows.")
    data_splitter: DataSplitter = Field(description="Controls the splitting of labeled data into training, validation, and test sets.")



"""
Sampler that chooses a random subset of annotation tasks.

Useful for debugging workflows end to end with a smaller
dataset.
"""

import random

from olmoearth_run.runner.models.training.labeled_data import AnnotationTask
from olmoearth_run.runner.tools.samplers.sampler_interface import SamplerInterface


class ChooseNSampler(SamplerInterface):
    """Sampler that chooses a random subset of annotation tasks."""

    def __init__(self, n: int):
        self.n = n

    def sample(self, annotation_tasks: list[AnnotationTask]) -> list[AnnotationTask]:
        """Choose a random subset of annotation tasks."""
        return random.sample(annotation_tasks, self.n)



import uuid
from datetime import datetime

from pydantic import BaseModel, Field

from olmoearth_shared.api.run.status import Status
from olmoearth_run.shared.models.api.step_type import StepType


class StepMetrics(BaseModel):
    total_compute_time_seconds: int | None = Field(default=None, description="The total compute time in seconds for this step. This is the sum of completed_at - started_at for all task attempts")


class StepResponse(BaseModel):
    """
    The model we expose over the API to describe a Step.
    Steps are typically provided as part of the WorkflowResponse
    """
    id: uuid.UUID
    workflow_id: uuid.UUID = Field(description="The ID of the Workflow this Step belongs to")
    step_type: StepType = Field(description="The type of Step, e.g., 'PREDICTION', 'DATASET_BUILD'")
    step_index: int = Field(ge=0, description="The index of the Step in the Workflow")
    status: Status = Field(description="The current status of the Step")
    metrics: StepMetrics = Field(description="Metrics related to the Step's execution")

    created_at: datetime
    updated_at: datetime
    started_at: datetime | None = Field(default=None, description="When the Step started running")
    completed_at: datetime | None = Field(
        default=None,
        description="The date and time when the Step was completed in either a success or failed state."
    )



from datetime import datetime
import uuid

from geojson_pydantic import Feature, FeatureCollection
from geojson_pydantic.geometries import Geometry, MultiPolygon, Polygon
from pydantic import BaseModel, Field, field_validator, model_validator


def _validate_time_range(start_time: datetime | None, end_time: datetime | None) -> None:
    if start_time is None or end_time is None:
        return

    if start_time > end_time:
        raise ValueError("Start time must be before end time")

    return


class AnnotationTaskFeatureProperties(BaseModel):
    """
    Properties for an Earth System Studio Annotation Task that contains a group of annotations.
    """
    oe_annotations_task_id: uuid.UUID = Field(description="The ID of the task that created the feature")
    oe_start_time: datetime = Field(description="The beginning of the temporal component")
    oe_end_time: datetime = Field(description="The end of the temporal component")

    @model_validator(mode='after')
    def validate_time_range(self) -> 'AnnotationTaskFeatureProperties':
        _validate_time_range(self.oe_start_time, self.oe_end_time)
        return self


class AnnotationFeatureProperties(BaseModel):
    oe_labels: dict[str, int | float | None]= Field(description="The labels of the feature (can have multiple fields for multi-task models)")
    oe_start_time: datetime | None = Field(default=None, description="The beginning of the temporal component")
    oe_end_time: datetime | None = Field(default=None, description="The end of the temporal component")
    oe_annotations_task_id: uuid.UUID = Field(description="The ID of the task that created the feature")

    @model_validator(mode='after')
    def validate_time_range(self) -> 'AnnotationFeatureProperties':
        _validate_time_range(self.oe_start_time, self.oe_end_time)
        return self

    @field_validator('oe_labels')
    @classmethod
    def validate_labels(cls, v: dict[str, int | float | None]) -> dict[str, int | float | None]:
        for value in v.values():
            if value is not None:
                return v

        raise ValueError("At least one label must be non-null")


class AnnotationTaskFeature(Feature[Polygon | MultiPolygon, AnnotationTaskFeatureProperties]):
    """A GeoJSON Feature for annotation tasks with guaranteed non-null geometry and properties."""

    # Override type annotations to be non-nullable
    geometry: Polygon | MultiPolygon
    properties: AnnotationTaskFeatureProperties

    def get_time_range(self) -> tuple[datetime, datetime]:
        """
        Extract temporal information from task properties.

        Returns:
            Tuple of (start_time, end_time)
        """
        return (self.properties.oe_start_time, self.properties.oe_end_time)


class AnnotationFeature(Feature[Geometry, AnnotationFeatureProperties]):
    """A GeoJSON Feature for annotations with guaranteed non-null geometry and properties."""

    # Override type annotations to be non-nullable
    geometry: Geometry
    properties: AnnotationFeatureProperties


class AnnotationTaskFeatureCollection(FeatureCollection[AnnotationTaskFeature]):
    """A GeoJSON FeatureCollection containing AnnotationTaskFeatures."""
    pass


class AnnotationFeatureCollection(FeatureCollection[AnnotationFeature]):
    """A GeoJSON FeatureCollection containing AnnotationFeatures."""
    pass



import logging
from datetime import timedelta
from http import HTTPStatus
from typing import Any

import requests
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry

logger = logging.getLogger(__name__)


class RequestSession:
    def __init__(
        self,
        total_retries: int,
        backoff_factor: float,
        backoff_max: timedelta,
        allowed_methods_for_retry: tuple[str, ...] = ("DELETE", "GET", "PATCH", "POST", "PUT"),
        statuses_for_retry: tuple[int, ...] = (
            HTTPStatus.TOO_MANY_REQUESTS,
            HTTPStatus.INTERNAL_SERVER_ERROR,
            HTTPStatus.BAD_GATEWAY,
            HTTPStatus.SERVICE_UNAVAILABLE,
            HTTPStatus.GATEWAY_TIMEOUT,
        ),
    ):
        self._session = requests.Session()
        retry = Retry(
            total=total_retries,
            backoff_factor=backoff_factor,
            backoff_max=backoff_max.total_seconds(),
            allowed_methods=allowed_methods_for_retry,
            status_forcelist=statuses_for_retry,
            raise_on_status=False,
            respect_retry_after_header=True,
        )
        self._session.mount("http://", HTTPAdapter(max_retries=retry))
        self._session.mount("https://", HTTPAdapter(max_retries=retry))
        self._session.hooks["response"] = lambda response, *args, **kwargs: response

    def request(self, method: str, url: str, **kwargs: Any) -> requests.Response:
        logger.debug("[request] %s %s %s", method, url, kwargs)
        return self._session.request(method, url, **kwargs)

    def delete(self, url: str, **kwargs: Any) -> requests.Response:
        return self.request("DELETE", url, **kwargs)

    def get(self, url: str, **kwargs: Any) -> requests.Response:
        return self.request("GET", url, **kwargs)

    def patch(self, url: str, **kwargs: Any) -> requests.Response:
        return self.request("PATCH", url, **kwargs)

    def post(self, url: str, **kwargs: Any) -> requests.Response:
        return self.request("POST", url, **kwargs)

    def put(self, url: str, **kwargs: Any) -> requests.Response:
        return self.request("PUT", url, **kwargs)



from typing import assert_never, cast

from olmoearth_run.runner.tools.data_splitters.data_splitter_interface import DataSplitterInterface
from olmoearth_run.runner.tools.data_splitters.random_data_splitter import RandomDataSplitter as RandomDataSplitterImpl
from olmoearth_run.runner.tools.data_splitters.spatial_data_splitter import SpatialDataSplitter as SpatialDataSplitterImpl
from olmoearth_run.runner.tools.labeled_window_preparers.labeled_window_preparer import LabeledWindowPreparer
from olmoearth_run.runner.tools.labeled_window_preparers.point_to_raster_window_preparer import PointToRasterWindowPreparer as PointToRasterWindowPreparerImpl
from olmoearth_run.runner.tools.labeled_window_preparers.polygon_to_raster_window_preparer import PolygonToRasterWindowPreparer as PolygonToRasterWindowPreparerImpl
from olmoearth_run.runner.tools.olmoearth_config.utils import ConfigUtils, OutputFieldUtils
from olmoearth_run.runner.tools.olmoearth_run_config_loader import OlmoEarthRunConfigLoader
from olmoearth_run.runner.tools.samplers.choose_n_sampler import ChooseNSampler as ChooseNSamplerImpl
from olmoearth_run.runner.tools.samplers.noop_sampler import NoopSampler as NoopSamplerImpl
from olmoearth_run.runner.tools.samplers.sampler_interface import SamplerInterface
from olmoearth_run.shared.models.olmoearth_run_config import OlmoEarthRunConfig
from olmoearth_shared.models.olmoearth_config.labeled_data_prep.data_splitter import RandomDataSplitter, SpatialDataSplitter
from olmoearth_shared.models.olmoearth_config.labeled_data_prep.samplers import NoopSampler, ChooseNSampler
from olmoearth_shared.models.olmoearth_config.labeled_data_prep.window_preparers import PointToRasterWindowPreparer, PolygonToRasterWindowPreparer
from olmoearth_shared.models.olmoearth_config.olmoearth_config import OlmoEarthConfig


class LabeledDataPrepClassLoader:
    @staticmethod
    def load_sampler(config: OlmoEarthRunConfig | OlmoEarthConfig) -> SamplerInterface:
        """Loads an annotation/task sampler instance based on provided config (legacy or unified)"""

        if isinstance(config, OlmoEarthRunConfig):
            return OlmoEarthRunConfigLoader.get_sampler(config)

        if not config.labeled_data_prep:
            raise ValueError("labeled_data_prep configuration is required for loading sampler")

        match config.labeled_data_prep.sampler:
            case NoopSampler():
                return NoopSamplerImpl()
            case ChooseNSampler(n=n):
                return ChooseNSamplerImpl(n=n)
            case unreachable:
                assert_never(unreachable)

    @staticmethod
    def load_data_splitter(config: OlmoEarthRunConfig | OlmoEarthConfig) -> DataSplitterInterface:
        """Loads a data splitter instance based on provided config (legacy or unified)."""
        if isinstance(config, OlmoEarthRunConfig):
            return OlmoEarthRunConfigLoader.get_data_splitter(config)

        if not config.labeled_data_prep:
            raise ValueError("labeled_data_prep configuration is required for loading data splitter")

        match config.labeled_data_prep.data_splitter:
            case RandomDataSplitter(seed=seed, train_prop=train_prop, val_prop=val_prop, test_prop=test_prop):
                return RandomDataSplitterImpl(seed=seed, train_prop=train_prop, val_prop=val_prop, test_prop=test_prop)
            case SpatialDataSplitter(grid_size=grid_size, train_prop=train_prop, val_prop=val_prop, test_prop=test_prop):
                return SpatialDataSplitterImpl(grid_size=grid_size, train_prop=train_prop, val_prop=val_prop, test_prop=test_prop)
            case unreachable:
                assert_never(unreachable)

    @staticmethod
    def load_labeled_window_preparer(config: OlmoEarthRunConfig | OlmoEarthConfig) -> LabeledWindowPreparer:
        """Loads a labeled window preparer instance based on provided config (legacy or unified)."""
        if isinstance(config, OlmoEarthRunConfig):
            return OlmoEarthRunConfigLoader.get_labeled_window_preparer(config)

        if not config.labeled_data_prep:
            raise ValueError("labeled_data_prep configuration is required for loading labeled window preparer")

        match config.labeled_data_prep.window_preparer:
            case PointToRasterWindowPreparer(window_buffer=window_buffer):
                output_field = ConfigUtils.get_raster_output_field(config)
                window_resolution = ConfigUtils.get_window_resolution(config)
                dtype = OutputFieldUtils.get_dtype_for_field(output_field)
                nodata_value = OutputFieldUtils.get_nodata_for_field(output_field)
                return PointToRasterWindowPreparerImpl(
                    window_buffer=window_buffer,
                    window_resolution=window_resolution,
                    dtype=dtype.value,
                    nodata_value=cast(int, nodata_value) if nodata_value is not None else 0,
                )
            case PolygonToRasterWindowPreparer():
                output_field = ConfigUtils.get_raster_output_field(config)
                window_resolution = ConfigUtils.get_window_resolution(config)
                nodata_value = OutputFieldUtils.get_nodata_for_field(output_field)
                return PolygonToRasterWindowPreparerImpl(
                    window_resolution=window_resolution,
                    nodata_value=cast(int, nodata_value) if nodata_value is not None else 0,
                )
            case unreachable:
                assert_never(unreachable)



from typing import Callable, TypeVar
from shapely.geometry import Polygon  #type: ignore[import-untyped]


T = TypeVar('T')


def group_polygons_by_grid(
    items: list[T],
    max_size: float,
    get_geometry: Callable[[T], Polygon]
) -> dict[tuple[int, int], list[T]]:
    """
    Group items with geometries into grid cells of maximum size.

    Each item is assigned to a grid cell based on its geometry's centroid.
    Grid cells are max_size x max_size in the coordinate system.

    Args:
        items: List of items (each must have a geometry accessible via get_geometry)
        max_size: Maximum width/height of each grid cell in coordinate units
        get_geometry: Function to extract Polygon from each item

    Returns:
        Dictionary mapping (grid_x, grid_y) to list of items in that cell
    """
    if not items:
        return {}

    # Find overall bounds to establish grid origin
    all_bounds = [get_geometry(item).bounds for item in items]
    min_x = min(b[0] for b in all_bounds)
    min_y = min(b[1] for b in all_bounds)

    # Group items by grid cell (based on centroid)
    from collections import defaultdict
    groups: dict[tuple[int, int], list[T]] = defaultdict(list)
    for item in items:
        geom = get_geometry(item)
        centroid = geom.centroid
        grid_x = int((centroid.x - min_x) // max_size)
        grid_y = int((centroid.y - min_y) // max_size)

        key = (grid_x, grid_y)
        groups[key].append(item)

    return groups



from pydantic import BaseModel, Field

from olmoearth_shared.models.olmoearth_config.model.encoders import Encoder
from olmoearth_shared.models.olmoearth_config.model.tasks import ModelTask


class Model(BaseModel):
    encoder: Encoder = Field(description="The encoder to use for the model. This produces embeddings from preprocessed modality data for use as input to task heads.")
    tasks: dict[str, ModelTask] = Field(description="The tasks to perform with the embeddings produced by the encoder layer. Each task will produce predictions for the original inputs.")



import logging
import multiprocessing
import os
from concurrent.futures import ThreadPoolExecutor

from rslearn.dataset import Dataset, Window
from rslearn.dataset.storage.file import FileWindowStorage, load_window
from upath import UPath

from olmoearth_run.config import OlmoEarthSettings
from olmoearth_run.runner.steps.base_step_definition import BaseStepDefinition
from olmoearth_run.runner.tools.dataset_builder import DatasetBuilder
from olmoearth_run.runner.tools.olmoearth_config.transpilers.olmoearth_to_rslearn import transpile_and_write_dataset_config
from olmoearth_run.shared.models.prediction_scratch_space import PredictionScratchSpace
from olmoearth_run.shared.models.api.step_type import StepType
from olmoearth_run.shared.models.api.task_args import DatasetBuildTaskArgs
from olmoearth_run.shared.models.api.task_results import DatasetBuildTaskResults

logger = logging.getLogger(__name__)

multiprocessing.set_start_method("forkserver", force=True)


class DatasetBuildStepDefinition(BaseStepDefinition[DatasetBuildTaskArgs, DatasetBuildTaskResults]):
    def run(self, task_args: DatasetBuildTaskArgs) -> DatasetBuildTaskResults:
        logger.info(f"Running dataset build step for {len(task_args.partition_ids)} partitions")
        scratch = PredictionScratchSpace(root_path=task_args.scratch_path)
        dataset_path = UPath(task_args.dataset_path)

        # JIT transpile OlmoEarthConfig to dataset config.json if using unified config
        olmoearth_config_path = UPath(scratch.olmoearth_config_path)
        if olmoearth_config_path.exists():
            transpile_and_write_dataset_config(scratch.olmoearth_config_path, dataset_path)

        # Get windows for partition
        dataset = Dataset(dataset_path)
        # Dataset building is a very IO intensive task, so we can have more workers than CPU cores
        builder = DatasetBuilder(
            dataset=dataset,
            num_workers=(os.cpu_count() or 1) * OlmoEarthSettings.DATASET_BUILD_WORKERS_PER_CPU,
            min_window_success_rate=task_args.min_window_success_rate
        )
        total_size_mb = 0.0

        for partition_id in task_args.partition_ids:
            logger.info(f"Building dataset for partition {partition_id}")
            partition_dir = UPath(task_args.dataset_path) / 'windows' / scratch.get_group(partition_id)
            window_paths = [window_dir for window_dir in partition_dir.iterdir() if window_dir.is_dir()]

            # Load windows in parallel using ThreadPoolExecutor for IO-bound operations
            with ThreadPoolExecutor() as executor:
                storage = FileWindowStorage(dataset.path)
                windows: list[Window] = list(executor.map(lambda wd: load_window(storage, wd), window_paths))

            builder.build_dataset(windows)
            total_size_mb += builder.get_partition_size_mb(str(partition_dir))

        return DatasetBuildTaskResults(
            step_type=StepType.DATASET_BUILD,
            dataset_build_path=task_args.dataset_path,
            dataset_size_mb=total_size_mb
        )
